(* generic-syntax.sml
 *
 * This module allows the client to specialize the syntax of the 
 * escape sequences to some degree via a callback.
 *
 * -- Allen Leung (leunga@{cs.nyu.edu, dorsai.org})
 *)

functor GenericRegExpSyntaxFn
  (structure R : GENERALIZED_REGEXP_SYNTAX

   (* Types of escape sequences *)
   datatype escape = 
     CHAR of R.Char.char             (* it's a character *)
   | MATCHSET of R.CharSet.set       (* it's a set *)
   | NONMATCHSET of R.CharSet.set    (* it's a set complemented *)
   | REGEXP of R.syntax              (* do the work in the client *)
   | CHARCODE of StringCvt.radix     (* do character code parsing *)
   | CTRL                            (* do control character parsing *)
   | REF of (string -> string) * int (* the ith back reference *)
   | ERROR of string                 (* it's an error! *)

   datatype context = IN_CHARSET | IN_REGEXP

   (* This callback is used to categorize escape sequences.
    * Escape sequences can appear in character sets and regular places
    *)
   type callbackdata

   val escape : callbackdata -> context -> 
                   (char, 'a) StringCvt.reader -> 
                   (escape, 'a) StringCvt.reader

   val dot : R.syntax (* what is . interpreted as *)

       (* Do we allow dangling modifiers such as /?/ /*/ /+/ ?
        * I.e. an empty string in front of these modifiers
        * If true, these will be treated as the empty string.
        * If false, there must something before these modifiers.
        *)
   val danglingModifiers : bool 

  ) : sig
          type callbackdata = callbackdata

          val scan : {data      : callbackdata, 
                      backslash : char, (* escape \ character *)
                      error     : 'a * string -> (R.syntax * 'a) option
                     } ->
                     (char, 'a) StringCvt.reader
	             -> (R.syntax, 'a) StringCvt.reader
      end =
struct

    structure R  = R
    structure RegExpSyntax = R
    structure S  = R.CharSet
    structure SC = StringCvt
    structure IS = IntRedBlackSet

    fun char c = R.Char.chr(Char.ord c)

    type callbackdata = callbackdata

    (* The syntax is LL(1) so there is no backtracking here! *)
    fun 'a scan {data, backslash, error} getc s =
    let exception ParseError of 'a * string

        fun err(s,msg) = raise ParseError(s,msg)

            (* the set of back references that appear in the regexp *)
        val backRefs = ref IS.empty 

        (* Can a pattern only match zero len strings? *)
        fun isZeroLen(R.Group x) = isZeroLen x
          | isZeroLen(R.Concat xs) = List.all isZeroLen xs
          | isZeroLen(R.Alt xs) = List.all isZeroLen xs
          | isZeroLen(R.Guard(_,x)) = isZeroLen x
          | isZeroLen(R.Star x) = isZeroLen x
          | isZeroLen(R.Plus x) = isZeroLen x
          | isZeroLen(R.Option x) = isZeroLen x
          | isZeroLen(R.Interval(_,_,SOME 0)) = true
          | isZeroLen(R.Interval(x,m,SOME n)) = m > n orelse isZeroLen x
          | isZeroLen(R.Interval(x,_,_)) = isZeroLen x
          | isZeroLen(R.Assign(_,_,x)) = isZeroLen x
          | isZeroLen(R.Begin) = true
          | isZeroLen(R.End) = true
          | isZeroLen(R.Boundary _) = true
          | isZeroLen _ = false

        fun concat [x] = x
          | concat stack  = R.Concat(rev stack)

        fun alt [x]   = x
          | alt stack = R.Alt(rev stack)

            (* Pull off a character from the stream *) 
        fun expect(ch, s) =
            case getc s of
              NONE => err(s, "missing "^Char.toString ch)
            | SOME(c',s') =>
              if ch = c' then s'
              else err(s, "expecting "^Char.toString ch^
                                       " but found "^Char.toString c')

        and dangling(s, c) = 
              if danglingModifiers then (s, [R.Concat []])
              else err(s, "dangling "^c)
 
        and star(s, stack as R.Star _::_) = (s, stack)
          | star(s, R.Plus x::stack) = (s, R.Star x::stack)
          | star(s, R.Interval(x,1,NONE)::stack) = (s, R.Star x::stack)
          | star(s, R.Interval(x,0,NONE)::stack) = (s, R.Star x::stack)
          | star(s, x::stack) = (s, (if isZeroLen x then
                                     R.Option x else R.Star x)::stack)
          | star(s, []) = dangling(s, "*")
         
        and plus(s, stack as R.Plus _::_) = (s, stack)
          | plus(s, stack as R.Star _::_) = (s, stack)
          | plus(s, R.Interval(x,1,NONE)::stack) = (s, R.Plus x::stack)
          | plus(s, stack as R.Interval(_,0,NONE)::_) = (s, stack)
          | plus(s, stack as x::xs) = (s, if isZeroLen x then stack
                                          else R.Plus x::xs)
          | plus(s, []) = dangling(s, "+")

        and opt(s, stack as R.Option _::_) = (s, stack)
          | opt(s, stack as R.Star _::_) = (s, stack)
          | opt(s, stack as R.Interval(_,0,_)::_) = (s, stack)
          | opt(s, x::stack) = (s, R.Option x::stack)
          | opt(s, []) = dangling(s, "?")

        and interval(s, stack) = 
            (* {n} 
             * {min,max}
             * {min,}
             *)
            let fun done(s, min, max) = 
                    case (min, max, stack) of 
                      (_, _, []) => dangling(s, "{}")
                    | (0, NONE, R.Option x::stack) => (s, R.Star x::stack)
                    | (0, NONE, stack as R.Star _::_) => (s, stack)
                    | (1, NONE, stack as R.Plus _::_) => (s, stack)
                    | (_, _, x::stack) => 
                         (s, if isZeroLen x then
                                if min > 0 then x::stack
                                else R.Option x::stack
                             else R.Interval(x, min, max)::stack)
                   
            in  case Int.scan SC.DEC getc s of  
                   SOME(min, s) => 
                   (case getc (SC.skipWS getc s) of
                     SOME(#",",s) => 
                     (case getc (SC.skipWS getc s) of
                       SOME(#"}",s) => done(s, min, NONE)
                     | SOME _ =>
                        (case Int.scan SC.DEC getc s of 
                          SOME(max, s) => 
                            done(expect(#"}", SC.skipWS getc s), min, SOME max)
                        | NONE => err(s, "illegal max in interval")
                        )
                     | NONE => err(s, "missing }")
                     )
                   | SOME(#"}",s) => done(s, min, SOME min)
                   | _ => err(s, "missing }")
                   )
                 | NONE => err(s, "missing min in interval")
            end

        (* Parse a character in a set *)
        and parseChar s =
            case getc s of
              NONE => err(s, "missing closing ]")
            | SOME(#"]",_) => NONE (* finished *)
            | SOME(#"[", _) => err(s, "dangling [ in set")
            | SOME(c,s') => (* handle escape sequences *)
              if c = backslash then
              (case escape data IN_CHARSET getc s' of
                 NONE => err(s, "dangling \\ in set")
               | SOME(ERROR msg, s') => 
                   err(s, "bad escape sequence in charset: "^msg)
               | SOME(CHARCODE radix, s') => 
                   let val (c, s') = parseCharCode(radix,s')
                   in  SOME(CHAR c, s')
                   end
               | SOME(CTRL, s) => let val (c, s) = parseCtrl s
                                  in  SOME(CHAR c, s)
                                  end
               | SOME(REGEXP(R.Char c),s) => SOME(CHAR c,s')
               | SOME(REGEXP(R.MatchSet set),s) => SOME(MATCHSET set,s')
               | SOME(REGEXP(R.NonmatchSet set),s) => SOME(NONMATCHSET set,s')
               | x => x
              ) 
              else SOME(CHAR(char c), s')

       (* Parse a character set expression *)
        and parseSet s = 
        let fun loop(s, set) = 
                case parseChar s of
                  NONE => (set, s)
                | SOME(CHAR c',s') => 
                  (case getc s' of
                    SOME(#"-",s'') => (* range? *)
                    (case parseChar s'' of 
                      NONE => err(s', "dangling -")
                    | SOME(CHAR c''',s''') => 
                         loop(s''', R.addRange(set,c',c'''))
                    | SOME(_, s') =>
                       err(s', "end range must be a character")
                    )
                  | _ => loop(s', S.add(set, c'))
                  ) 
                | SOME(MATCHSET set',s) =>
                   loop(s, S.union(set,set'))
                | SOME(NONMATCHSET set',s) =>
                   loop(s, S.union(set,S.difference(R.allChars, set')))
                | SOME _ => 
                   err(s, "bug in charset escape sequence")
            val (mode, (set, s)) = 
                case getc s of
                  SOME(#"^", s') => 
                    (R.NonmatchSet, loop(s', S.empty))
                | _  => (R.MatchSet, loop(s, S.empty))
        in  (mode set, s)
        end
    

        (* the main parser *)
        and parse(s, stack) =
            (case getc s of
              NONE => (concat stack, s)
            | SOME(c',s') =>
              (case c' of
                #")" => (concat stack, s)
              | #"|" => (concat stack, s)
              | #"(" => let val (re, s') = parseAlt s'
                        in  parse(expect(#")", s'), R.Group re::stack)
                        end
              | #"[" => let val (re, s') = parseSet s'
                        in  parse(expect(#"]", s'), re::stack)
                        end
              | #"^" => parse(s', R.Begin::stack)
              | #"$" => parse(s', R.End::stack)
              | #"." => parse(s', dot::stack)
              | #"*" => parse(star(s', stack))
              | #"+" => parse(plus(s', stack))
              | #"?" => parse(opt(s', stack))
              | #"{" => parse(interval(s', stack))
              | #"]" => err(s, "dangling ]")
              | #"}" => err(s, "dangling }")
              | c' =>
                if c' = backslash then
                   let val (re, s') = parseEscape s'
                   in  parse(s', re::stack)
                   end
                else parse(s', R.Char(char c')::stack)
              )
            )

        and parseAlt s = 
            let fun loop(s, alts) =
                    case getc s of 
                      NONE => (alt alts, s)
                    | SOME(#"|",s') => 
                      let val (re, s'') = parse(s', []) 
                      in  loop(s'', re::alts)
                      end
                    | SOME _ => (alt alts, s)
                val (re, s) = parse(s, [])
            in  loop(s, [re])
            end
  
          (* do character code parsing *)
        and parseCharCode(radix, s) =
            case Int.scan radix getc s of
              SOME(n, s) =>  
                ((R.Char.chr n, s) 
                   handle Chr =>
                       err(s, "character code out of range"))
            | NONE => err(s, "bad character code")

          (* do control character parsing *)
        and parseCtrl s =
            case getc s of
              NONE => err(s, "bad control character")
            | SOME(#"[",s) => (R.Char.chr 27,s)
            | SOME(c,s) =>
              if Char.isAlpha c then
                 (R.Char.chr(Char.ord(Char.toLower c) - Char.ord #"a" + 1), s)
              else err(s, "bad control character")
  
          (* do escape sequence *)
        and parseEscape s =
            (case escape data IN_REGEXP getc s of
              NONE => err(s, "dangling \\")
            | SOME(CHAR c,s)         => (R.Char c, s)
            | SOME(CHARCODE radix,s) => let val (c, s) = parseCharCode(radix, s)
                                        in (R.Char(c), s) 
                                        end
            | SOME(MATCHSET set,s)     => (R.MatchSet set, s)
            | SOME(NONMATCHSET set, s) => (R.NonmatchSet set, s)
            | SOME(CTRL, s) => let val (c, s) = parseCtrl s
                               in  (R.Char(c), s)
                               end
            | SOME(REGEXP re,s)  => (re, s)
            | SOME(REF(f,i), s)  => 
                  (backRefs := IS.add(!backRefs, i);
                   (R.Ref(f, i), s)
                  )
            | SOME(ERROR msg,s)  => 
                err(s, "bad escape sequence "^msg)
            )

          (* do postprocessing if backreferences are used *)  
        fun walk(i, R.Concat xs) = walkList R.Concat (i, xs)
          | walk(i, R.Alt xs)  = walkList R.Alt (i, xs)
          | walk(i, R.Group x) = 
            let val (j, x) = enter R.Group (i+1, x) (* new group *)
            in  (j, if IS.member(!backRefs, i) 
                    then R.Assign(i,fn x => x,x) else x)
            end
          | walk(i, R.Star x) = enter R.Star (i, x)
          | walk(i, R.Plus x) = enter R.Plus (i, x)
          | walk(i, R.Option x) = enter R.Option (i, x)
          | walk(i, R.Guard(p, x)) = enter (fn x => R.Guard(p,x)) (i, x)
          | walk(i, R.Interval(x,a,b)) = 
              enter (fn x => R.Interval(x,a,b)) (i, x)
          | walk(i, x as R.Ref _) = (i, x)
          | walk(i, x as R.MatchSet _) = (i, x)
          | walk(i, x as R.NonmatchSet _) = (i, x)
          | walk(i, x as R.Char _) = (i, x)
          | walk(i, x as R.Boundary _) = (i, x)
          | walk(i, x as R.Begin) = (i, x)
          | walk(i, x as R.End) = (i, x)
          | walk(i, R.Assign _) = err(s, "bug!!!")

        and walkList f (i, xs) = 
            let fun loop(i, []) = (i, [])
                  | loop(i, x::xs) = 
                    let val (i, x) = walk(i, x)
                        val (i, xs) = loop(i, xs) 
                    in  (i, x::xs)
                    end
                val (i, xs) = loop(i, xs)
            in  (i, f xs) 
            end

        and enter f (i, x) = 
            let val (j, x) = walk(i, x)
            in  (j, f x)
            end

        fun scanIt s = 
        let val (re, s) = parseAlt s 
            val re = if IS.isEmpty(!backRefs) then re else #2(walk(1,re))
        in  SOME(re, s)
        end
    in  scanIt s handle ParseError(s,msg) => error(s,msg)
    end

end (* GenericSyntax *)
