(* perl-syntax.sml
  
   This module implements a subset of PERL regular expressions syntax.  

   WARNING: there is no locale support!
  
   The meta characters are:
     "\" "^" "$" "." "[" "]" "|" "(" ")" "*" "+" "?"

           \   Quote the next metacharacter
           ^   Match the beginning of the line
           .   Match any character (except newline)
           $   Match the end of the line (or before newline at the end)
           |   Alternation
           ()  Grouping
           []  Character class


       The following standard quantifiers are recognized:

           *      Match 0 or more times
           +      Match 1 or more times
           ?      Match 1 or 0 times
           {n}    Match exactly n times
           {n,}   Match at least n times
           {n,m}  Match at least n but not more than m times

          \033        octal char (think of a PDP-11)
           \x1B        hex char
           \x{263a}    wide hex char         (Unicode SMILEY)
           \c[         control char
           \N{name}    named char
           \l          lowercase next char (think vi)
           \u          uppercase next char (think vi)
           \L          lowercase till \E (think vi)
           \U          uppercase till \E (think vi)
           \E          end case modification (think vi)
           \Q          quote (disable) pattern metacharacters till \E

           \pP Match P, named property.  Use \p{Prop} for longer names.
           \PP Match non-P
           \C  Match a single C char (octet) even under utf8.

               Perl defines the following zero-width assertions:

           \b  Match a word boundary
           \B  Match a non-(word boundary)
           \A  Match only at beginning of string
           \Z  Match only at end of string, or before newline at the end
           \z  Match only at end of string
 *)

functor PerlSyntaxFn(R : GENERALIZED_REGEXP_SYNTAX) : 
sig 
   include GENERALIZED_REGEXP_PARSER where R = R

   (* With user supplied error handler *)
   val scan': ('a * string -> (R.syntax * 'a) option) ->
               (char,'a) StringCvt.reader ->
                (R.syntax,'a) StringCvt.reader
end =
struct
  structure R = R

  structure S = 
  GenericRegExpSyntaxFn
  (structure R = R
   structure S = R.CharSet

   fun char c = R.Char.chr(ord c)

   fun mkSet s = S.addList(S.empty, map char (String.explode s))

   val danglingModifiers = false  (* don't allow things like /?/ *)

   datatype escape = 
     CHAR of R.Char.char
   | MATCHSET of S.set
   | NONMATCHSET of S.set
   | REGEXP of R.syntax 
   | CHARCODE of StringCvt.radix 
   | CTRL 
   | REF of (string -> string) * int
   | ERROR of string            

   datatype context = IN_CHARSET | IN_REGEXP


    (*
           \t          tab                   (HT, TAB)
           \n          newline               (LF, NL)
           \r          return                (CR)
           \f          form feed             (FF)
           \a          alarm (bell)          (BEL)
           \e          escape (think troff)  (ESC)
     *)
 
   val tab = CHAR(char #"\t")
   val nl  = CHAR(char #"\n")
   val cr  = CHAR(char #"\r")
   val ff  = CHAR(char #"\f")
   val bel = CHAR(char #"\a")
   val esc = CHAR(char #"\027") (* NOTE: SML uses decimal! not octal *)

   (* IMPORTANT NOTE: perl's . also matches \000 *)
   val dot = R.NonmatchSet (mkSet "\n")

   (* What perl means: 
    *
    *)
   fun isWord c = R.Char.isAlphaNum c orelse R.Char.ord c = 95

   (*
    *      \w  Match a "word" character (alphanumeric plus "_")
    *      \W  Match a non-word character
    *      \s  Match a whitespace character
    *      \S  Match a non-whitespace character
    *      \d  Match a digit character
    *      \D  Match a non-digit character
    *)

   fun every p = 
   let fun iter(i,S) = 
           let val S = if p i then i::S else S
           in  if R.Char.>=(i,R.Char.maxChar) then S else iter(R.Char.succ i, S)
           end
   in  S.addList(S.empty, iter(R.Char.minChar, [])) 
   end


   val words       = R.addRange(S.empty,char #"a",char #"z")
   val words       = R.addRange(words,char #"A",char #"Z")
   val words       = R.addRange(words,char #"0",char #"9")
   val words       = S.add(words,char #"_")
   val word        = MATCHSET words
   val nonword     = NONMATCHSET words

   val spaces      = every R.Char.isSpace
   val space       = MATCHSET spaces
   val nonspace    = NONMATCHSET spaces

   val digits      = R.addRange(S.empty,char #"0",char #"9")
   val digit       = MATCHSET digits
   val nondigit    = NONMATCHSET digits

   val hex         = CHARCODE StringCvt.HEX
   val oct         = CHARCODE StringCvt.OCT

   (* Perl defines the following zero-width assertions:
    *
    *      \b  Match a word boundary
    *      \B  Match a non-(word boundary)
    *      \A  Match only at beginning of string
    *      \Z  Match only at end of string, or before newline at the end
    *      \z  Match only at end of string
    *)
              (* word at begining of string *)
   fun isWordBoundary{prev=NONE, curr=SOME c, next} = isWord c 
              (* word at end of string *)
     | isWordBoundary{prev=SOME c,curr=NONE, next}  = isWord c 
     | isWordBoundary{prev=NONE, curr=NONE, next}   = false   (* empty string *)
     | isWordBoundary{prev=SOME x,curr=SOME y, next} = 
         if isWord x then not(isWord y) else isWord y 

   fun beginOfString{prev=NONE, curr, next} = true
     | beginOfString _ = false

   fun endOfString{curr=NONE, prev, next} = true
     | endOfString _ = false

   fun endOfString'{curr=NONE, prev, next} = true
     | endOfString'{curr=SOME c, next=NONE, ...} = R.Char.ord c = 10
     | endOfString' _ = false

   val wordB    = REGEXP(R.Boundary isWordBoundary)
   val nonwordB = REGEXP(R.Boundary(not o isWordBoundary))
   val beginB   = REGEXP(R.Boundary beginOfString)
   val endB     = REGEXP(R.Boundary endOfString)
   val endB'    = REGEXP(R.Boundary endOfString')

   type callbackdata = unit

   (* Handle quoting of
    *   \Q ... \E 
    *   \U ... \E      to upper case
    *   \L ... \E      to lower case
    * All meta characters are treated as normal within these.
    * I think the semantics here is the same as perl's.
    *)
   fun quote transform context getc s = 
   let fun loop(s, chars) =
           case getc s of
             NONE => SOME(ERROR "missing \\E", s)
           | SOME(c as #"\\",s) =>
             (case getc s of
               SOME(#"E",s) => done(chars, s) 
             | _ => loop(s, char(transform c)::chars)
             )
           | SOME(c, s) => loop(s, char(transform c)::chars)

       and done(chars, s) = 
           case (context, chars) of
             (_, [c]) => SOME(CHAR c, s)
           | (IN_REGEXP, cs) => SOME(REGEXP(R.Concat(rev(map R.Char cs))), s)
           | (IN_CHARSET, cs) =>  SOME(MATCHSET(S.addList(S.empty, cs)), s)
   in  loop(s, [])
   end

   (* Callback for escape sequences *)
   fun escape data context getc s = 
       case (getc s, context) of
         (NONE, _) => NONE
 
         (* Simple escapes *)
       | (SOME(#"t",s), _) => SOME(tab,s)
       | (SOME(#"n",s), _) => SOME(nl,s)
       | (SOME(#"r",s), _) => SOME(cr,s)
       | (SOME(#"f",s), _) => SOME(ff,s)
       | (SOME(#"a",s), _) => SOME(bel,s)
       | (SOME(#"e",s), _) => SOME(esc,s)

         (* Character codes; no unicode support yet! *)
       | (SOME(#"0",s), _) => SOME(oct,s)
       | (SOME(#"x",s), _) => SOME(hex,s)
       | (SOME(#"c",s), _) => SOME(CTRL,s)

         (* Not yet supported *)
       | (SOME(#"N",s), _) => SOME(ERROR "named character",s)
       | (SOME(#"l",s), _) => SOME(ERROR "lowercase next char",s)
       | (SOME(#"u",s), _) => SOME(ERROR "uppercase next char",s)

         (* character set abbreviations *)
       | (SOME(#"w",s), _) => SOME(word,s)
       | (SOME(#"W",s), _) => SOME(nonword,s)
       | (SOME(#"s",s), _) => SOME(space,s)
       | (SOME(#"S",s), _) => SOME(nonspace,s)
       | (SOME(#"d",s), _) => SOME(digit,s)
       | (SOME(#"D",s), _) => SOME(nondigit,s)

         (* Quoting *) 
       | (SOME(#"Q",s), _) => quote (fn x=>x) context getc s
       | (SOME(#"L",s), _) => quote Char.toLower context getc s
       | (SOME(#"U",s), _) => quote Char.toUpper context getc s

         (* Boundary operators; these cannot appear in a char set *)
       | (SOME(#"b",s), IN_REGEXP) => SOME(wordB, s)
       | (SOME(#"B",s), IN_REGEXP) => SOME(nonwordB, s)
       | (SOME(#"A",s), IN_REGEXP) => SOME(beginB, s)
       | (SOME(#"Z",s), IN_REGEXP) => SOME(endB', s)
       | (SOME(#"z",s), IN_REGEXP) => SOME(endB, s)

         (* Property *) 
         (* What are these?
       | (SOME(#"p",s), _) => (SOME PROPERTY,s)
       | (SOME(#"P",s), _) => (SOME NONPROPERTY,s)
          *)

       | (SOME(c, s'), _)  => 
          if Char.isDigit c andalso context = IN_REGEXP
          then (* It's a back reference 
                * BUG: Int.scan is too greedy.  
                * How to handle things like \11 where 
                *    is it \1 and the character 1?
                *)
             (case Int.scan StringCvt.DEC getc s of
                SOME(i, s) => SOME(REF(fn x => x,i), s)
             |  NONE => SOME(ERROR "back reference error",s)
             )
             (* By default just treat the character literally *)
          else SOME(CHAR(char c), s')
  )


  fun scan' err getc = S.scan{data=(), backslash= #"\\", error=err} getc

  fun scan getc = scan' (fn _ => raise R.CannotParse) getc
end

structure PerlSyntax = PerlSyntaxFn(RegExpSyntax)
