(*
 * This module is used to generate state tables with the following
 * encoding scheme:
 *    delta(s,c) = 
 *      let val offset = base[s] + c
 *      in  if check[offset] = s then next[offset] else error
 *      end
 * where base, check, and next are three 1-D arrays.
 *
 * This encoding scheme is suitable for "sparse tables", i.e., tables with
 * only a few non-error transitions out of each state.   LALR(1) tables
 * are usually sparse.
 *
 * The compression heuristic is very simple:
 *   1. If the new state has only one non-error transition symbol 
 *         (singleton states), save it later.
 *   2. Pack the normal states with first fit.
 *   3. When all non-singleton states are dealt with, pack the singleton states
 *      into the "holes" found in the table.
 *)

local

   val DEBUG = false

in

structure SparseAutomataTable :> AUTOMATA_TABLE =
struct

   val error = 0
  
   val i2s = Int.toString

   exception AutomataTable of string
   structure A = Array
   structure DynArray = 
   struct
      datatype 'a array = ARR of 'a A.array ref * int ref * 'a
 
      fun sub(ARR(ref a,_,d),i) = if i < A.length a then A.sub(a,i) else d

      fun setBound(ARR(arr' as ref a,b,d),n) =
          if n >= A.length a then
          let val a' = A.array(n+1, d)
          in  A.copy{src=a,dst=a',si=0,di=0,len=SOME(!b+1)};
              arr' := a';
              b    := n
          end
          else b := Int.max(n,!b)

      fun update(arr as ARR(ref a,b,d),i,x) = 
          let val n = A.length a 
          in  if i >= n then 
                  (setBound(arr, Int.max(8,Int.max(n+n,i+1)));
                   update(arr, i, x)
                  )
              else (A.update(a, i, x);
                    if !b < i then b := i else ()
                   )
          end 
      fun length(ARR(_,ref b,_)) = b+1
      fun array(n,x) = ARR(ref(A.array(n,x)),ref ~1,x)
      fun toList(ARR(ref a,ref b,_)) = 
          A.foldri (fn (_,x,l) => x::l) [] (a,0,SOME(b+1))
      fun toArray(ARR(ref a,_,_)) = a
      fun toPacked(ARR(ref a,ref b,_), w) =
      let fun pack(x, 0, s) = s
            | pack(x, i, s) =
              let val b = Word.toIntX(Word.andb(x,0wxff))
              in  pack(Word.>>(x,0wx8), i-1, Char.chr b::s)
              end
          val l = A.foldri(fn (_,x,s) => pack(Word.fromInt x,w,s)) [] 
                     (a,0,SOME(b+1))
      in  String.implode l
      end
   end
   structure DA = DynArray

   type state = int
   type symbol = int
   datatype table = 
      TABLE of 
      {base       : int DA.array,
       check      : state DA.array,
       next       : state DA.array,
       singletons : (state * symbol * state) list ref,
       minOffset  : int ref,
       maxOffset  : int ref,
       minChar    : symbol,
       maxChar    : symbol,
       maxState   : state ref
      }

   fun min(TABLE{minChar, ...}) = minChar
   fun max(TABLE{maxChar, ...}) = maxChar

   fun stateCount(TABLE{maxState,...}) = !maxState + 1

   fun expand(TABLE{minChar, maxChar, check, next, ...}, offset) = 
   let val offset = offset + maxChar - minChar 
   in  DA.setBound(check, offset);
       DA.setBound(next, offset)
   end 
 
   fun addState (table as TABLE{singletons, base, check, next, maxState, 
                                minOffset, maxOffset, minChar, maxChar, ...}) 
                (state, delta) =
       (maxState := Int.max(!maxState,state);
        DA.update(base,state,0);
        case delta of
          [] => () (* done *)
        | [(c,s)] => singletons := (state,c,s) :: !singletons
        |  _ =>
          (* Fit into the table using a first fit strategy *)       
          let fun scan i =
                  if DA.sub(check, i) = error then i else scan(i+1)
              val n = scan(!minOffset)
              val _ = minOffset := n
              fun fits(i, []) = true
                | fits(i, (c,s)::delta) =
                  DA.sub(check, i + c) = error andalso fits(i, delta) 
              fun find(i) = if fits(i, delta) then i else find(i+1)
              val offset = find (n - minChar) (* base offset *)
              val _ = if DEBUG then print("Offset="^i2s offset^"\n") else ()
          in  
              app (fn (c,s) =>
                       (DA.update(check, offset + c, state);
                        DA.update(next, offset + c, s))) delta;
              DA.update(base, state, offset);
              if !maxOffset < offset then maxOffset := offset else ();
              expand(table,!maxOffset)
          end
       )

   fun new{min,max} = 
   let val minSize = max-min+1
       val table = TABLE{base=DA.array(1,0),
                         check=DA.array(minSize,error),
                         next=DA.array(minSize,error),
                         singletons=ref [],
                         minChar=min,
                         maxChar=max,
                         maxState=ref error,
                         minOffset=ref 1,
                         maxOffset=ref 0
                        }
   in  table
   end

   fun charRange(TABLE{minChar, maxChar, ...}) = {min=minChar, max=maxChar}

       (* Insert singleton transitions into the table *)
   fun compress (table as TABLE{singletons, base, check, next, 
                              minChar, minOffset, maxOffset, maxChar, ...}) = 
       let fun insert(i, []) = i
             | insert(i, trans as (s,c,s')::rest) =
               if DA.sub(check, i) = error then (* hole found *)
                  let val offset = i-c
                  in  if offset < 0 then insert(c, trans)
                      else
                      (DA.update(check, i, s);
                       DA.update(next, i, s');
                       DA.update(base, s, offset);
                       if !maxOffset < offset then maxOffset := offset else ();
                       insert(i+1, rest)
                      )
                  end
               else 
                  insert(i+1, trans)

             (* lowest character first *)
            val toDo = ListMergeSort.sort(fn ((_,c,_),(_,c',_)) =>c>c')
                                (!singletons)
            val _ = insert(!minOffset, !singletons) 
        in  singletons := [];
            expand(table,!maxOffset)
        end

   fun go (TABLE{singletons=ref [], base, check, next, ...}) =
       let val base = DA.toArray base
           val check =DA.toArray check
           val next = DA.toArray next
           fun delta(state, c) =
           let val offset = A.sub(base, state) + c
           in  if A.sub(check, offset) = state 
               then A.sub(next, offset) else error
           end
       in  delta
       end
     | go table = (compress table; go table)  (* compile it first *)

   fun getrep (TABLE{base, check, next, ...}) rep = 
       (case rep of
         "base"  => base
       | "check" => check
       | "next"  => next
       | rep     => raise AutomataTable("unknown table "^rep)
       )

   fun widthOf n = if n < 256 then 1 else if n < 65536 then 2 else 4

   fun getwidth(TABLE{maxState, maxOffset, ...}) rep = 
       (case rep of
         "base"  => widthOf(!maxOffset)
       | "check" => widthOf(!maxState)
       | "next"  => widthOf(!maxState)
       | rep     => raise AutomataTable("unknown table "^rep)
       )

   fun getarr table rep = DA.toArray(getrep table rep)

   fun getpacked table rep = 
   let val w = getwidth table rep
   in  (DA.toPacked(getrep table rep, w), w)
   end

end

end
