(******************************************************************************)
theory Lueng_regexp_lib_sig_and_sml
imports Pure "Lueng"
begin
(******************************************************************************)

ML{*
(*
 * A lazy man's interface to the regexp library.
 *)
signature REGEXP_LIB =
sig
   type regexp = string
   type text   = string

        (* search *)
   val grep  : regexp -> text -> string list option

        (* find all occurrances matching regexp *)
   val findAll : regexp -> text -> string list

        (* find all occurrances of the ith group in the matching regexp *)
   val findAllGroup : regexp -> int -> text -> string list

        (* match and extract the match; return "" otherwise *)
   val extract : regexp -> text -> string

        (* match and extract the ith group; return "" otherwise *)
   val extractGroup : regexp -> int -> text -> string

        (* search only *)
   val matches  : regexp -> text -> bool

        (* search and replace *) 
   val subst : regexp -> (string list -> string) -> text -> text

        (* search and replace all occurrances *) 
   val substAll : regexp -> (string list -> string) -> text -> text

        (* search and replace *) 
   val replace : regexp -> string -> text -> text

        (* search and replace all occurrances *) 
   val replaceAll : regexp -> string -> text -> text

        (* multiway match *)
   val match : text ->
               { cases: (regexp * (string list -> 'a)) list,
                 default: unit -> 'a
               } -> 'a
end
*}

ML{*
(*
 * A lazy man's interface to the regexp library.
 *)
structure RegExpLib :> REGEXP_LIB =
struct

   structure RE = RegExpFn(structure P = PerlSyntax
                           structure E = PerlEngine
                          )

   structure M  = MatchTree

   (* For caching compiled regexp *)
   structure H = HashTableFn(type hash_key = string
                             val hashVal = HashString.hashString
                             val sameKey = op= : string * string -> bool
                            )

   type regexp = string and text = string


   val cache = H.mkTable(16,Match) : RE.regexp H.hash_table

   fun compile regexp =
       case H.find cache regexp of
         SOME re => re
       | NONE =>
         let val re = RE.compileString regexp
         in  H.insert cache (regexp, re);
             re
         end

   fun search regexp text =
       StringCvt.scanString (RE.find (compile regexp)) text

   fun getArgs text children = 
   let fun walk (M.Match(SOME{pos,len},children)) = 
           let val s = String.substring(text,pos,len)
           in  s::List.concat(map walk children)
           end
         | walk (M.Match(NONE,children)) = ""::List.concat(map walk children)
   in List.concat(map walk children) end

   fun grep regexp text = 
       case search regexp text of
         NONE => NONE
       | SOME(M.Match(_,children)) => SOME(getArgs text children)

   fun extractGroup regexp i text =
       case search regexp text of
         NONE => ""
       | SOME m => (case M.nth(m,i) of 
                      NONE => ""
                    | SOME{pos,len} => String.substring(text,pos,len)
                   ) handle _ => ""

   fun extract regexp = extractGroup regexp 0

   fun look regexp text = 
   let val n    = size text
       fun getc i = if i >= n then NONE else SOME(String.sub(text,i), i+1)
   in  RE.find(compile regexp) getc
   end

   fun findAllGroup regexp g text = 
   let val look = look regexp text
       fun loop s = 
           case look s of
             SOME(m,s) => 
               (case M.nth(m,g) of
                 SOME{pos,len} => String.substring(text,pos,len)::loop s
               | NONE => loop s
               )
           | NONE => [] 
   in  loop 0
   end

   fun findAll regexp = findAllGroup regexp 0 

   fun matches regexp text = Option.isSome(search regexp text) 

   fun match text {cases,default} = 
   let fun loop [] = default()
         | loop((regexp,action)::rest) = 
           case grep regexp text of
             NONE => loop rest
           | SOME args => action args
   in  loop cases end

   fun subst regexp f text = 
       case search regexp text of
         NONE => text
       | SOME(M.Match(SOME{pos,len},children)) =>
         let val prefix = String.extract(text,0,SOME pos)
             val suffix = String.extract(text,pos+len,NONE)
         in  prefix^f(getArgs text children)^suffix
         end
       | SOME _ => raise Subscript

   fun substAll regexp f text = 
   let val look = look regexp text
       fun loop s =
           case look s of
             NONE => [if s = 0 then text else String.extract(text,s,NONE)]
           | SOME(M.Match(SOME{pos,len},children),s') =>
             let val prefix = String.substring(text,s,pos-s)
             in  prefix::f(getArgs text children)::loop s'
             end
           | SOME _ => raise Subscript
   in  String.concat(loop 0)
   end

   fun replace regexp s = subst regexp (fn _ => s) 
   fun replaceAll regexp s = substAll regexp (fn _ => s) 

end
*}


(******************************************************************************)
end
(******************************************************************************)






