(* perl-engine.sml
 *
 * Implements a perl-like regular expressions matcher. 
 * This module is based on backtracking search.
 *
 * TODO: 
 *   1. Compile suitable subexpressions into DFA 
 *   2. Lookahead optimizations when scanning for substring match.
 * 
 * --- Allen Leung (leunga@{cs.nyu.edu,dorsai.org})
 *)

local 
   (* Set this to true if you need a representation that is thread-safe; 
    * When this is true the backreferences table will be allocated anew
    * every time a new match is performed.
    *)
   val threadSafe = true
in

functor PerlEngineFn(R : REGEXP_SYNTAX) : PERL_REGEXP_ENGINE =
struct

   structure R = R
   structure S = R
   structure D = DfaEngine
   structure M = MatchTree
   structure A = Array

   datatype regexp =
     REGEXP of 
       {refCount  : int,             (* number of backreference variables *)
        references: string A.array,  (* references table *)
        regexp    : S.syntax,
        multiline : bool,            (* multiline match? *)
        beginOnly : bool,            (* can only match at beginning of line? *)
        lookahead : S.syntax option  (* a simple lookahead test *) 
       }

   exception BackTrack
   exception Bug

   fun multiline(REGEXP{refCount, references, regexp, beginOnly, 
                        lookahead, ...}) = 
       REGEXP{refCount=refCount, references=references, regexp=regexp,
              beginOnly=beginOnly, lookahead=lookahead, multiline=true}

   fun singleline(REGEXP{refCount, references, regexp, beginOnly, 
                        lookahead, ...}) = 
       REGEXP{refCount=refCount, references=references, regexp=regexp,
              beginOnly=beginOnly, lookahead=lookahead, multiline=false}

   fun optimize re = re (* not implemented yet *)

   fun collectInfo re =
   let             (* results means: match begin only, must be empty *)
       val n = ref ~1
       fun updt v = if v > !n then n := v else ()
       fun f(S.Char _) = (false, false)
         | f(S.MatchSet _) = (false, false)
         | f(S.NonmatchSet _) = (false, false)
         | f(S.Plus re) = f re
         | f(S.Star re) = f re
         | f(S.Option re) = f re
         | f(S.Begin) = (true, true)
         | f(S.End) = (false, true)
         | f(S.Group re) = f re
         | f(S.Concat es) = g es
         | f(S.Alt es) = h es
         | f(S.Interval(re,min,SOME 0)) = (false, true)
         | f(S.Interval(re,min,max)) = f re
         | f(S.Guard(_,re)) = f re
         | f(S.Assign(v,_,re)) = (updt v; f re)
         | f(S.Ref(f,v)) = (updt v; (false, false))
         | f(S.Boundary _) = (false, true)
       and g [] = (false, true)
         | g (e::es) = 
           let val (beginOnly, isEmpty) = f e
               val (beginOnly', isEmpty') = g es
           in  (beginOnly orelse isEmpty andalso beginOnly',
                isEmpty andalso isEmpty')
           end
       and h [] = (false, false) (* can never match anything *)
         | h [e] = f e
         | h (e::es) = 
           let val (beginOnly, isEmpty) = f e
               val (beginOnly', isEmpty') = h es
           in  (beginOnly andalso beginOnly', 
                isEmpty andalso isEmpty'
               ) 
           end
       val (beginOnly, isEmpty) = f re
   in  (beginOnly, isEmpty, !n+1) 
   end

   fun compile re = 
   let val (beginOnly, isEmpty, refCount) = collectInfo re
   in  REGEXP{regexp=re, 
              refCount=refCount,
              multiline=false,
              beginOnly=beginOnly,
              lookahead=NONE,
              references=if threadSafe then A.array(0,"")
                         else A.array(refCount, "")
             }
   end

   fun scan(REGEXP{regexp, multiline, ...}, references,
            getc, startPos, stream, last) =
   let 
       fun lesseq(i, NONE) = true
         | lesseq(i, SOME max) = i <= max
       fun less(i, NONE) = true
         | less(i, SOME max) = i < max

       (* Get n characters from the stream.
        * There must be at least n characters in the stream.
        *)
       fun getVal(n, s) = 
       let fun f(0, s) = []
             | f(n, s) = 
               case getc s of
                 SOME(c, s) => c::f(n-1, s)
               | NONE => raise Bug
       in  String.implode(f(n, s))
       end

       (* Get and set references *)
       fun deref v = A.sub(references,v)
       fun assign(v, s) = A.update(references, v, s)
 
       (* At the beginning of the line? *)
       fun isBeginning(0, last) = true
         | isBeginning(_, NONE) = true
         | isBeginning(_, SOME(#"\n",_)) = multiline
         | isBeginning(_, _) = false

       (* This function gets an empty match structure, for when the appropriate
        * alternative is not followed at all
        *)
       fun emptyAlt (S.Group e) = [M.Match (NONE,emptyAlt e)]
         | emptyAlt (S.Alt l) = List.concat (map emptyAlt l)
         | emptyAlt (S.Concat l) = List.concat (map emptyAlt l)
         | emptyAlt (S.Interval (e,_,_)) = emptyAlt e
         | emptyAlt (S.Option e) = emptyAlt e
         | emptyAlt (S.Star e) = emptyAlt e
         | emptyAlt (S.Plus e) = emptyAlt e
         | emptyAlt (S.Assign(v, _, e)) = 
              (assign(v,""); (* reset reference *)
               emptyAlt e) 
         | emptyAlt (S.Guard(_, e)) = emptyAlt e
         | emptyAlt (_) = []


       (* The arguments are:
        *
        *     re           = regexp list
        *     pos          = current integer position of the stream
        *     peek         = one character lookahead buffer
        *     last         = buffer of last scanned character 
        *     s            = stream
        *     ms           = match tree stack (NOTE: in reversed order)
        *     K            = continuation
        *
        * "last" is necessary to implement various kinds of boundary like
        * operators.
        *)
       datatype 'a cont = 
           GROUP of S.syntax list * 'a * int * 'a match_tree list * 'a cont
         | ASSIGN of int * (string -> string) *
                     S.syntax list * 'a * int * 'a match_tree list * 'a cont
         | GUARD of (string -> bool) *
                     S.syntax list * 'a * int * 'a match_tree list * 'a cont
         | CONCAT of S.syntax list * 'a cont
         | REPEAT of S.syntax * int * int option * int * 
                     S.syntax list * 'a match_tree list * 'a cont
         | RETURN

       withtype 'a match_tree = 
                 {pos : 'a, len : int} option MatchTree.match_tree
       and 'a buf = (char * 'a) option

       (* interp and cont are tail recursive except for backtracking *)
       fun cont(pos, peek, last, s, nested, GROUP(re, s', pos', siblings, K)) = 
           let val ms = M.Match(SOME{pos=s', len=pos-pos'}, rev nested)
                           ::siblings
           in  interp(re, pos, peek, last, s, ms, K)
           end
         | cont(pos, peek, last, s, ms2, ASSIGN(v, f, re, s', pos', ms1, K)) = 
           let val oldValue = deref v
               val ms = ms2 @ ms1
           in  assign(v, f (getVal(pos - pos', s')));
               interp(re, pos, peek, last, s, ms, K)
                   (* when backtracking; reset value of backreference *)
                 handle BackTrack => (assign(v, oldValue); raise BackTrack)
           end
         | cont(pos, peek, last, s, ms2, GUARD(pred, re, s', pos', ms1, K)) =  
           let val ms = ms2 @ ms1
           in  if pred (getVal(pos - pos', s')) then
                  interp(re, pos, peek, last, s, ms, K)
               else
                  raise BackTrack
           end
         | cont(pos, peek, last, s, ms, CONCAT(re, K)) =
             interp(re, pos, peek, last, s, ms, K)

         | cont(pos, peek, last, s, ms2, REPEAT(x, min, max, ith, re, ms1, K)) =
               if ith < min then
                  interp([x], pos, peek, last, s, [],
                         REPEAT(x, min, max, ith+1, re, ms1, K))
               else if less(ith, max) then
                  (interp([x], pos, peek, last, s, [],
                         REPEAT(x, min, max, ith+1, re, ms1, K))
                   handle BackTrack => 
                     interp(re, pos, peek, last, s, ms2@ms1, K)
                  )
               else 
                  interp(re, pos, peek, last, s, ms2@ms1, K)
         | cont(pos, peek, last, s, ms, RETURN) = (pos, peek, last, s, rev ms)

       and interp([], pos, peek, last, s, ms, K) = 
              cont(pos, peek, last, s, ms, K)
         | interp(S.Char c::re, pos, peek as SOME(c',s'), last, s, ms, K) = 
              if c = c' 
              then interp(re, pos+1, NONE, peek, s', ms, K)
              else raise BackTrack
         | interp(S.MatchSet set::re,pos, peek as SOME(c',s'), last, s, ms, K) =
              if S.CharSet.member(set, c') 
              then interp(re, pos+1, NONE, peek, s', ms, K)
              else raise BackTrack
         | interp(S.NonmatchSet set::re, pos, peek as SOME(c',s'),last,s,ms,K) =
              if S.CharSet.member(set, c') 
              then raise BackTrack
              else interp(re, pos+1, NONE, peek, s', ms, K)
         | interp(S.Group x::re, pos, peek, last, s, ms, K) =
              interp([x], pos, peek, last, s, [], GROUP(re, s, pos, ms, K))
         | interp(S.Alt []::_, pos, peek, last, s, ms, K) = raise BackTrack
         | interp(S.Alt [x]::re, pos, peek, last, s, ms, K) =
              interp(x::re, pos, peek, last, s, ms, K)
         | interp(S.Alt(x::xs)::re, pos, peek, last, s, ms, K) = 
              (interp(x::re, pos, peek, last, s, ms, K)
               handle BackTrack =>
               interp(S.Alt xs::re, pos, peek, last, s, ms, K)
              )
         | interp(S.Concat x::re, pos, peek, last, s, ms, K) = 
              interp(x, pos, peek, last, s, ms, CONCAT(re, K)) 
         | interp(S.Star x::re, pos, peek, last, s, ms, K) =
              interp(S.Interval(x,0,NONE)::re, pos, peek, last, s, ms, K) 
         | interp(S.Plus x::re, pos, peek, last, s, ms, K) =
              interp(S.Interval(x,1,NONE)::re, pos, peek, last, s, ms, K)
         | interp(S.Option x::re, pos, peek, last, s, ms, K) = 
              interp(S.Interval(x,0,SOME 1)::re, pos, peek, last, s, ms, K)

         | interp(S.Interval(x,min,max)::re, pos, peek, last, s, ms, K) =
           let fun emptyMatch() = 
                   interp(re, pos, peek, last, s, 
                          List.revAppend(emptyAlt x,ms), K)
               fun tryAtLeastOne() = 
                   interp([x], pos, peek, last, s, [], 
                          REPEAT(x, min, max, 1, re, ms, K))
           in  if lesseq(min,max) then 
                   if min > 0 then tryAtLeastOne()
                   else (tryAtLeastOne() handle BackTrack => emptyMatch())
               else (* the range is empty *)
                   emptyMatch()
           end

         | interp(S.Begin::re, pos, peek, last, s, ms, K) = 
              if isBeginning(pos, last) 
              then interp(re, pos, peek, last, s, ms, K)
              else raise BackTrack

         | interp(S.End::re, pos, peek, last, s, ms, K) =
              (case getc s of
                 NONE => interp(re, pos, peek, last, s, ms, K)
               | last' as SOME(c',s') => 
                   if multiline andalso c' = #"\n" then
                      interp(re, pos+1, NONE, last', s', ms, K)
                   else raise BackTrack
              )
           (* Assign a variable *)
         | interp(S.Assign(v, f, x)::re, pos, peek, last, s, ms, K) =
              interp([x], pos, peek, last, s, [], ASSIGN(v,f,re,s,pos,ms,K))
           (* Match a back reference *)
         | interp(S.Ref(f,v)::re, pos, peek, last, s, ms, K) = 
              let val text = f(deref v)
                  val n = size text
                  fun matches(i, last, s) = 
                      if i >= n then 
                         interp(re, pos+n, NONE, last, s, ms, K)
                      else case getc s of
                             last as SOME(c',s') => 
                               if c' = String.sub(text,i) 
                               then matches(i+1, last, s')
                               else raise BackTrack
                          |  NONE => raise BackTrack
              in  matches(0, last, s)
              end 

          | interp(S.Guard(pred, x)::re, pos, peek, last, s, ms, K) =
              interp([x], pos, peek, last, s, [], GUARD(pred,re,s,pos,ms,K))
              
           (* Generalized boundary operator *)
         | interp(S.Boundary ok::re, pos, peek, last, s, ms, K) =
           let val prev = case last of SOME(c,_) => SOME c | NONE => NONE
               val (curr,next) = 
                    case getc s of 
                      SOME(c,s') => 
                      (case getc s' of
                        NONE => (SOME c, NONE)
                      | SOME(c',_) => (SOME c,SOME c')
                      )
                    | NONE => (NONE, NONE)
           in  if ok{prev=prev,curr=curr,next=next} then
                  interp(re, pos, peek, last, s, ms, K)
               else
                  raise BackTrack
           end

         | interp(re, pos, NONE, last, s, ms, K) =  (* peek ahead *)
            (case getc s of
               NONE => raise BackTrack
             | peek => interp(re, pos, peek, NONE, s, ms, K)
            ) 

       val (pos, peek, last, s, matches) = 
           interp([regexp], startPos, NONE, last, stream, [], RETURN) 
   in  SOME(M.Match(SOME{pos=stream, len=pos-startPos},matches), s)
   end handle BackTrack => NONE

   fun allocRefs(REGEXP{refCount, references, ...}) = 
       if threadSafe andalso refCount > 0 
       then A.array(refCount, "") else references

   fun prefix regexp getc stream = 
          scan(regexp, allocRefs regexp, getc, 0, stream, NONE) 

   fun find (regexp as REGEXP{beginOnly, multiline, ...}) getc stream =  
   let val refs = allocRefs regexp
  
       (* Most general and slowest *)
       fun loop1(pos, stream, last) = 
           case scan(regexp, refs, getc, pos, stream, last) of
              NONE =>
               (case getc stream of
                  NONE => NONE
               |  last as SOME(c,s) => loop1(pos+1, s, last)
               )
           |  x => x

       (* Multiline pattern that only matches at the beginning of line *)
       fun loop2(pos, stream, last) = 
           case scan(regexp, refs, getc, pos, stream, last) of
              NONE => skipToNextLine(pos, stream)
           |  x => x

       and skipToNextLine(pos, stream) =
           case getc stream of
              NONE => NONE
           |  last as SOME(#"\n",s) => loop2(pos+1, s, last)
           |  last as SOME(_,s) => skipToNextLine(pos+1, s)

   in  case (beginOnly, multiline) of
          (true, false) => scan(regexp, refs, getc, 0, stream, NONE)
        | (false, _) => loop1(0, stream, NONE) 
        | (true, true) => loop2(0, stream, NONE)
   end

   (* Executes the longest match *)
   fun match rules = 
   let (* precompile all patterns *)
       val rules = map (fn (re,act) => (compile re, act)) rules
       fun doit getc stream = 
       let fun findBest([], NONE) = NONE
             | findBest([], SOME(len, match, s, act)) = SOME(act match, s)
             | findBest((re,act)::rest, best) = 
               (case (prefix re getc stream, best) of
                  (NONE, best) => findBest(rest, best)
                | (SOME(m as M.Match(SOME{len, ...},_), s), NONE) => 
                   findBest(rest, SOME(len, m, s, act))
                | (SOME(m as M.Match(SOME{len, ...},_), s), 
                   SOME(len', _, _, _)) =>
                   if len > len' then findBest(rest, SOME(len, m, s, act))
                   else findBest(rest, best)
                |  _ => raise Bug
               )
       in  findBest(rules, NONE)
       end
   in  doit
   end

end

structure PerlEngine = PerlEngineFn(RegExpSyntax)

end (* local *)
