(*
 * Test the PerlSyntax and PerlEngine modules 
 *)

CM.make "../regexp-lib.cm";

local 
   structure RE = RegExpFn(structure P = PerlSyntax
                           structure E = PerlEngine
                          )

   structure M  = MatchTree

   fun search regexp = StringCvt.scanString (RE.find (RE.compileString regexp)) 

   fun getArgs text children = 
   let fun walk (M.Match(SOME{pos,len},children)) = 
           let val s = String.substring(text,pos,len)
           in  s::List.concat(map walk children)
           end
         | walk (M.Match(NONE,children)) = ""::List.concat(map walk children)
   in List.concat(map walk children) end


   fun pr x = "["^pr' x^"]"
   and pr' [] = ""
     | pr' [x] = "\""^x^"\""
     | pr' (x::xs) = "\""^x^"\","^pr' xs
  
   fun show s = (print s; ())

in
   fun grep regexp text = 
       case search regexp text of
         NONE => NONE
       | SOME(M.Match(_,children)) => SOME(getArgs text children)

   fun matches regexp text expected =
       case grep regexp text of
          SOME result => 
           if result = expected then
             show("OK /"^regexp^"/ matches \""^text^"\" result="^pr result^"\n")
           else
             show("BUG /"^regexp^"/ matches \""^text^"\" result="^pr result^
                  " expected="^pr expected^"\n")
       | NONE =>
          show("BUG /"^regexp^"/ fails to match \""^text^
               "\" expected="^pr expected^"\n")

   fun fails regexp text =
       case grep regexp text of 
         SOME result =>
          show ("BUG /"^regexp^"/ should not match \""^text^
                "\" result="^pr result^"\n")
       | NONE => 
          show ("OK /"^regexp^"/ fails to match \""^text^"\" as expected\n")

   fun demo() =
       ( matches "a(.*)b" "baabcdefbsf" ["abcdef"];
         matches "a(.+)b" "cabbab" ["bba"];
         matches "(\\w+)\\s*=\\s*(\\d+)" "number = 123" ["number","123"];
         matches "(\\w+)\\s*=\\s*(-?\\d+(\\.\\d*)?)" "real = -123.23"
                 ["real", "-123.23", ".23"];
         matches "(\\w+)\\s*=\\s*(-?\\d+(\\.\\d*)?)" "real=-123";
                 ["real", "-123", ""];
         matches "a(.*)b(.*)c" "ccdcdcabcdfdsae" ["",""];
         fails "^abc" "acdcdabc";
         fails "abc$" "acbcabcdc";
         matches "a*(aba)*"  "aaabaabaaba" [""];
         matches "a*(ab\\w)*$" "aaabaabbabc" ["abc"];
         matches "^[a-z]$" "a" [];
         matches "^[A-Z]$" "A" [];
         matches "\\Q.$\\][)(}{\\E" ".$\\][)(}{" []; (* Testing \Q and \E *)
         matches "\\bAllen\\b" "Allen Leung" []; 
         fails "\\bAllen\\b" "AllenLeung"; 
         matches "(a{5,7})" "aaaaaaaaaa" ["aaaaaaa"]; 
         matches "(a{5,5})" "baaaaaaa" ["aaaaa"]; 
         matches "(a{5,3})" "aaaaaaa" [""]; 
         matches "((ab){3,3})" "fffffababababababab" ["ababab","ab"]; 
         matches "(\\d{3,})" "01xxx02xx0234xx" ["0234"];  
         matches "^\\ca$" "\001" [];
         matches "^\\c[$" "\027" [];
         matches "^(x{3}){3}$" "xxxxxxxxx" ["xxx"];
         matches "^(()*)*" "xxxxxxxxx" ["",""];
         matches "\\[(([^\\]]|\\.)*)\\]" "[a-zA-Z0-9]" ["a-zA-Z0-9","9"];
         matches "[^0-9]" "a" [];
         matches "#x([0-9a-hA-H]+)" "#xa" ["a"];
         fails "^(x{3}){3}$" "xxxxxxxx";
         fails "^(x{3}){3}$" "xxxxxxxxxx";
         app
           (fn x => 
              matches "^(while|for|do|structure|functor|end|case|fun|val)$"
                x [x])
               ["fun", "while", "case", "do"];
         app
           (fails "^(while|for|do|structure|functor|end|case|fun|val)$")
               ["typedef", "goto", "switch", "return"];

         ()
       )

end
