(* syntax-sig.sml
 *
 * COPYRIGHT (c) 1995 AT&T Bell Laboratories.
 *
 * This is the abstract syntax tree used to represent regular expressions.
 * It serves as the glue between different front-ends (implementing
 * different RE specification languages), and different back-ends (implementing
 * different compilation/searching algorithms).
 *)
(*GC44 CHANGES
   Converted "signature GENERALIZED_REGEXP_SYNTAX" 
   to the normal SML/NJ "REGEXP_SYNTAX"
*)
signature REGEXP_SYNTAX = 
  sig

      exception CannotParse
      exception CannotCompile

          (* Parameterized by the character representation 
           * This generalization is for unicode support.
           *)
      structure Char : CHAR  

      structure CharSet : ORD_SET where type Key.ord_key = Char.char

      datatype syntax
        = Group of syntax
	| Alt of syntax list
	| Concat of syntax list
	| Interval of (syntax * int * int option)
	| Option of syntax	(* == Interval(re, 0, SOME 1) *)
	| Star of syntax	(* == Interval(re, 0, NONE) *)
	| Plus of syntax	(* == Interval(re, 1, NONE) *)
	| MatchSet of CharSet.set
	| NonmatchSet of CharSet.set
	| Char of Char.char
	| Begin                   (* Matches beginning of stream *)
	| End                     (* Matches end of stream *)

          (* Extensions *)
        | Assign of int * (string -> string) * syntax  
             (* define a reference *)
        | Ref of (string -> string) * int (* back references *)
        | Guard of (string -> bool) * syntax
        | Boundary of {prev: Char.char option,
                       curr: Char.char option,
                       next: Char.char option} -> bool

      val addRange : CharSet.set * Char.char * Char.char -> CharSet.set
      val allChars : CharSet.set
	  
  end


(* 
 * Some notes on the extensions:
 *
 * What's Assign and Ref?
 * ------------------------
 *
 * They are used to implement perl-like back references:
 * For example, the perl regexp:
 *
 *   /(.+)\1/
 *
 * is compiled into the syntax
 * 
 *   Concat(Assign(1,fn x => x,Group(Plus allChars)),Ref(fn x => x,1))
 *
 * This matches repeated strings like:
 *
 *    xyzxyz
 *    abab
 *
 * Optionally, Ref and Assign can apply a function on the back reference.
 * For example, define
 * 
 *   fun revString s = String.implode o rev o String.explode
 *
 * Then the syntax:
 *
 *   Concat(Assign(1,fn s => s,Group(Plus allChars)),
 *          Ref(SOME(String.implode o rev o String.explode),1))
 *
 * matches palindromes like:
 *
 *     xyzzyx 
 *     abba
 *
 * What is Guard?
 * --------------
 *    Guard allows an arbitrary predicate to be attached to a regexp.
 *
 * For example, 
 *  
 *   Concat(Assign(1,Group(Plus allChars)),
 *          Guard(Char.contains #"x" (Ref(fn x => x, 1))))
 *
 * Matches repeated strings like
 * 
 *    xyzxyz
 *
 * but not
 *
 *    abab
 *
 * because the character #x has to appear in the back reference 
 *)
