(*
 * This module is used to generate state tables with the following encode
 * scheme:
 *
 *   delta(s,c) = 
 *      let val offset = base[s] + c
 *      in  if check[offset] = s then next[offset] else delta(default[s],c)
 *      end
 *  where base, check, next, and default are all 1-D arrays
 *
 * This encoding scheme can be used to encode "dense" state tables with
 * a lot of similar states.   Lexer tables are usually dense.
 * 
 *)
local
   val DEBUG = false

   (* a maximum of this many templates at any one time *)
   val maxTemplates = 12

   (* A default state can only have this many differences *)
   fun maxDiffsWithDefaultState numChars = 
       if (numChars <= 32) then numChars div 3 
       else if (numChars <= 64) then numChars div 4
       else numChars div 5

   (* Create a new template whenever this many differences are encountered *)
   fun minAddTemplateDiffs numChars = numChars div 3

in
structure DenseAutomataTable :> AUTOMATA_TABLE =
struct
   structure S  = SparseAutomataTable 
   structure DA = S.DynArray
   structure A  = Array
   open S

   type time = int

   datatype template = 
       TEMPLATE of 
          { state   : state,
            uses    : int ref,
            lastUse : time ref,
            delta   : (symbol * state) list
          }

   datatype table = 
        TABLE of
         { default   : state DA.array,
           sparse    : S.table,
           clock     : time ref,
           templates : template list ref
         }

   val i2s = Int.toString

   fun new{min,max} = 
   let val t = S.new{min=min,max=max}
   in  TABLE{default=DA.array(1,error), sparse=t, templates=ref [], clock=ref 0}
   end

   fun trans2s tr = 
         "["^String.concat(map (fn (c,s) => i2s c^"->"^i2s s^" ") tr)^"]"

   fun stateCount(TABLE{sparse, ...}) = S.stateCount sparse

   fun charRange(TABLE{sparse, ...}) = S.charRange sparse

   fun addState(TABLE{sparse, default, templates, clock, ...}) =
   let val {min, max} = S.charRange sparse
       val numChars = max-min+1
       val minRatio = minAddTemplateDiffs numChars 
       val maxDiffs = maxDiffsWithDefaultState(numChars)
       val _ = clock := !clock + 1
 
       val sortDelta = ListMergeSort.sort 
                         (fn ((c:symbol,s:state),(c',s')) => c>c')

       val addState = S.addState sparse

       fun add(s, delta) = 
       let val delta = sortDelta delta
           val fanOut = length delta

           (* Add a new state *)
           fun newState(true, 
                        SOME(TEMPLATE{state=t, lastUse, uses,...}, n, diffs)) = 
               (if DEBUG then
                  print("Defaulted state "^i2s s^" default="^i2s t^" diffs="^
                        trans2s diffs^"\n") else ();
                addState (s, diffs);
                DA.update(default, s, t);
                lastUse := !clock;
                uses := !uses + 1
               )
             | newState _ = 
               (if DEBUG then print("Sparse state "^i2s s^ 
                                    " delta="^trans2s delta^"\n") else ();
                addState (s, delta);
                DA.update(default, s, error)
               )

           (* Compare two transition list.
            *    Return the number of differences and 
            *    the transition differences 
            *)
           fun cmp([], [], n, diffs) = (n, rev diffs)
             | cmp(xs, [], n, diffs) = (n + length xs, rev diffs@xs)
             | cmp([], ys, n, diffs) = 
                 (n + length ys, rev diffs@map (fn (c,_) => (c,error)) ys)
             | cmp(a as (tr as (c:symbol,s:state))::xs, 
                   b as (c',s')::ys, n, diffs) = 
               if c = c' then 
                  if s = s' then cmp(xs, ys, n, diffs)
                  else cmp(xs, ys,n+1,tr::diffs)
               else if c < c' then
                  cmp(xs, b, n+1, tr::diffs)
               else
                  cmp(a, ys, n+1, (c',error)::diffs)

           (* Look for the best match *)
           fun bestMatch([], best) = best
             | bestMatch((t as TEMPLATE{delta=delta', ...})::rest, best) =
               let val (n, diffs) = cmp(delta, delta', 0, []) 
                   val best = 
                       case best of
                         NONE => SOME(t, n, diffs)
                       | SOME(_,n',_) => 
                          if n < n' then SOME(t, n, diffs) else best
               in  bestMatch(rest, best)
               end

           (* Remove a template *)
           fun prune(templates) =
           let fun findVictim([], victim) = victim
                 | findVictim((t as TEMPLATE{uses, lastUse, ...})::rest, 
                              victim) =
                   case victim of 
                     NONE => SOME t
                   | SOME(TEMPLATE{uses=uses', lastUse=lastUse', ...}) =>
                     findVictim(rest, 
                        if !uses < !uses' orelse
                           !uses = !uses' andalso
                           !lastUse < !lastUse' then SOME t else victim)
           in  case findVictim(templates, NONE) of
                 NONE => templates
               | SOME(TEMPLATE{state=v, delta,...}) => 
                  (if DEBUG then 
                     print("Removing template "^i2s v^
                           " delta="^trans2s delta^"\n") else ();
                   List.filter (fn TEMPLATE{state,...} => state<>v) templates
                  )
           end

           (* Add a new template *)
           fun addTemplate() = 
           let 
               val ts = TEMPLATE{state=s, delta=delta, uses=ref 1, 
                                 lastUse=ref(!clock)} :: !templates
           in  if DEBUG then print("New template "^i2s s^
                                   " delta="^trans2s delta^"\n") else ();
               if length ts > maxTemplates then
                  templates := prune ts
               else templates := ts
           end

           val template = bestMatch(!templates, NONE)

           (* Should we use the found template as default? *)
           val usingTemplate =
                case template of
                  NONE => false
                | SOME(_,n,_) => n < fanOut andalso n < maxDiffs 
                     (* only if the differences are not too great *)

           (* Should we add the current state also as a new template? *)
           val addingTemplate = 
               fanOut > 1 andalso
               (not usingTemplate orelse
                (case template of 
                   NONE => false
                 | SOME(TEMPLATE{uses, ...},n,_) =>
                    n > 1 andalso !uses > 10 orelse n > minRatio
                )
               )

       in  newState(usingTemplate, template);
           if addingTemplate then addTemplate() else ()
       end
   in  add
   end

   fun compress(TABLE{sparse, ...}) = S.compress sparse

   fun getrep(TABLE{default, ...}) "default" = default
     | getrep(TABLE{sparse, ...}) rep = S.getrep sparse rep

   fun getarr table rep = DynArray.toArray(getrep table rep)

   fun getwidth (TABLE{sparse, ...}) "default" = S.getwidth sparse "base"
     | getwidth (TABLE{sparse, ...}) rep = S.getwidth sparse rep

   fun getpacked table rep = 
   let val w = getwidth table rep
   in  (DynArray.toPacked(getrep table rep, w), w)
   end

   fun go(table as TABLE{default, ...}) =
   let val base    = getarr table "base"
       val check   = getarr table "check"
       val next    = getarr table "next"
       val default = getarr table "default"
       fun delta(s, c) = 
       let val offset = A.sub(base,s) + c
       in  if A.sub(check,offset) = s 
           then A.sub(next,offset) 
           else let val s = A.sub(default, s)
                in  if s = error then error else delta(s, c)
                end
       end
   in  delta
   end 

end
end 
