(*
 * Yet another dfa based pattern matching engine.
 * 
 * Features: 
 *  1. FastLexerEngine is a bit faster than DfaEngine in dfa construction.
 *  2. The metacharacter ^ and $ are supported.
 *     However, backreferences and other zero-length assertions like \b 
 *     are unsupported.
 *
 * This version is specialized to the usual 8-bit characters.
 * Note that the full match tree is not reported.
 * For that, you need LexerEngine.
 *)
structure FastLexerEngine : REGEXP_ENGINE =
struct

   structure R     = RegExpSyntax 
   structure Char  = R.Char
   structure Gen   = LexerGenerator(R)
   structure Lexer = LexerTables
   structure M     = MatchTree
   structure A     = Array

   type regexp = Gen.Lexer.lexer

   fun compile re = Gen.compile [re]

   val error = 0 (* the error state is always 0 *)

   (* Create a new scanner *) 
   fun scanner(lexer as 
                  Lexer.LEXER{equiv, base, check, next, skels, numGroups, 
                              default, accept, eof, 
                              start, start', ...},
               prefixOnly
              ) =
   let (* Generate the match tree from a successful match *) 
       fun extract(rule, stream, endPos) = 
           M.Match(SOME{pos=stream, len=endPos}, [])

       (* This function decodes the transition tables *)
       fun go(s,c) =
       let val offset = A.sub(base,s) + c
       in  if A.sub(check,offset) = s 
           then A.sub(next,offset)
           else let val s' = A.sub(default,s) 
                in  if s' = error then error else go(s',c)
                end
       end

       val state0 = A.sub(start, 0)
       val state0' = A.sub(start', 0)
       val rule0 = A.sub(accept, state0)
       val rule0' = A.sub(accept, state0')

       fun doScan getc stream = 
       let fun scan(s0, r0, pos, stream) = 
           let 
               fun loop(state, pos, s, rule_ok, pos_ok, s_ok) =
                   case getc s of
                     NONE => endOfStream(state, pos, s, rule_ok, pos_ok, s_ok)
                   | SOME(c,s') =>
                     let val state' = go(state, A.sub(equiv,Char.ord c))
                     in  if state' = error 
                         then finish(rule_ok, pos_ok, s_ok) 
                         else  
                         let val rule = A.sub(accept,state')
                         in  if rule >= 0 then (* An accept state *)
                                loop(state', pos+1, s', rule, pos, s)
                             else
                                loop(state', pos+1, s', rule_ok, pos_ok, s_ok)
                         end
                     end
               and finish(~1, pos, s) = 
                    if prefixOnly then NONE 
                    else if state0 = error then fastcontinue(pos, stream)
                         else continue(pos, stream)
                 | finish(rule, pos, s) = 
                    SOME((extract(rule,stream,pos+1), rule), s)
    
               and endOfStream(state, pos, s, rule_ok, pos_ok, s_ok) =
                   let val state' = go(state,eof)
                   in  if state' = error 
                       then finish(rule_ok, pos_ok, s_ok)
                       else 
                       let val rule = A.sub(accept,state')
                       in  if rule >= 0 then (* An accept state *)
                              finish(rule, pos, s)
                           else
                              finish(rule_ok, pos_ok, s_ok)
                       end
                  end
    
               and continue(pos, s) = 
                   case getc s of 
                     NONE => NONE
                   | SOME(#"\n",s) => scan(state0', rule0', pos+1, s)
                   | SOME(_,s) => scan(state0, rule0, pos+1, s) 

               and fastcontinue(pos, s) = 
                   case getc s of 
                     NONE => NONE
                   | SOME(#"\n",s) => scan(state0',rule0', pos+1, s)
                   | SOME(_,s) => fastcontinue(pos+1, s) 
           
           in  loop(s0, pos, stream, r0, ~1, stream)
           end
       in  scan(state0', rule0', 0, stream)
       end
    
   in  doScan 
   end

   fun prefixOrFind(lexer,prefixOnly) = 
   let val scanner = scanner(lexer, prefixOnly)
       fun scan getc stream =
           case scanner getc stream of
             SOME((m, _), s) => SOME(m, s)
           | NONE => NONE
   in  scan
   end

   fun prefix lexer = prefixOrFind(lexer,true)
   fun find   lexer = prefixOrFind(lexer,false)

   fun match rules =
   let val lexer = Gen.compile (map (fn (re,_) => re) rules)
       val actions = Vector.fromList (map #2 rules)
       val scanner = scanner(lexer, true)
       fun matcher getc stream =
           case scanner getc stream of
              SOME((m, ruleno), s) => SOME((Vector.sub(actions,ruleno)) m,s)
           |  NONE                 => NONE
   in  matcher
   end

end
