local

   val DEBUG = false

in

structure FSA :> FSA =
struct

   structure S        = IntRedBlackSet
   structure H        = IntHashTable
   structure M        = IntRedBlackMap
   structure A        = Array
   structure IntSet   = S

   type symbol  = int
   type ruleno  = int
   type groupid = int

   exception FSA

   datatype node = NODE of {id:int, kind:kind ref}
   and      kind = Accept of ruleno
                 | Epsilon of node 
                 | Or of node list
                 | Delta of symbol * node 
                 | BeginGroup of groupid * node
                 | EndGroup of groupid * node
                 | Undef
                 | Error

   datatype nfa = 
      NFA of
      {  minChar  : symbol,
         maxChar  : symbol,
         nextid   : int ref,
         nodes    : node H.hash_table,
         closure  : S.set H.hash_table
      }



   val i2s = Int.toString
   fun intsetToString s = 
        "{"^S.foldr(fn (i,"") => i2s i | (i,s) => i2s i^","^s) "" s^"}"
   val s2s = intsetToString

   fun nid(NODE{id, ...}) = "<"^i2s id^">"

   fun n2s(node as NODE{kind=ref kind, ...}) =
       nid node^
       (case kind of
          Accept r  => " accept "^i2s r^"\n"
        | Epsilon x => " --> "^nid x^"\n"
        | Or xs => " --> "^foldr (fn (x,"") => nid x | 
                                     (x,s) => nid x^","^s) "" xs^"\n"
        | Delta(c, x) => " -- "^Char.toString(chr c)^" --> "^nid x^"\n"
        | BeginGroup(i, x) => " ("^i2s i^" --> "^nid x^"\n"
        | EndGroup(i, x) => " )"^i2s i^" --> "^nid x^"\n"
        | Undef => " undef\n"
        | Error => " error\n"
       )

   fun nfa{min, max} = NFA{minChar=min, maxChar=max, 
                           nodes=H.mkTable(32,FSA), 
                           closure=H.mkTable(32,FSA),
                           nextid=ref 0}

   fun newNode(NFA{nodes, nextid, ...}) = 
   let val id = !nextid
       val _ =  nextid := id + 1
       val node = NODE{id=id, kind=ref Undef}
   in  H.insert nodes (id,node); node
   end

   fun setAccept(n as NODE{kind, ...}, r) = (kind := Accept r; n)
   fun setEpsilon(n as NODE{kind, ...}, x) = (kind := Epsilon x; n)
   fun setOr(n as NODE{kind, ...}, ns) = (kind := Or ns; n)
   fun setDelta(n as NODE{kind, ...}, c, x) = (kind := Delta(c, x); n)
   fun setBeginGroup(n as NODE{kind, ...}, i, x) = (kind := BeginGroup(i, x); n)
   fun setEndGroup(n as NODE{kind, ...}, i, x) = (kind := EndGroup(i, x); n)
   fun setError(n as NODE{kind, ...}) = (kind := Error; n)
  
   fun computeClosure(NFA{nodes, closure, nextid, ...}) =
   let 
       val insert = H.insert closure
       val find   = H.find closure
       fun lookup n = getOpt(find n,S.empty)

       (* A table of nodes to is dependents *)
       val dependents = H.mkTable (16,FSA)
       val findDep    = H.find dependents
       val insertDep  = H.insert dependents
       fun lookupDep n = getOpt(findDep n,[]) 

       val onWorkList = A.array(!nextid,false)
       fun insertWL(x,WL) =
           if A.sub(onWorkList,x) then WL
           else (A.update(onWorkList,x,true); x::WL)

           (* x depends on y; i.e. there is an epsilon edge x --> y*)
       fun addEdge(x,y) = 
            (if DEBUG then print(i2s x^" --> "^i2s y^"\n") else ();
             insertDep(y,x::lookupDep y)
            )

       (* Iterate until fix point; dumb *)
       fun iter [] = ()
         | iter(x::WL) = 
           let val ys = lookupDep x
               val Sx = lookup x
               fun propagate([], WL) = iter WL
                 | propagate(y::ys, WL) = 
                   if x = y then propagate(ys, WL)
                   else let val Sy = lookup y
                            val ny = S.numItems Sy
                            val Sy = S.union(Sx,Sy)
                        in  if S.numItems Sy > ny then 
                               (insert(y, Sy);
                                if DEBUG then
                                   print("Closure "^i2s y^"="^s2s Sy^"\n")
                                else ();
                                propagate(ys, insertWL(y,WL))
                               )
                            else propagate(ys, WL)
                        end
           in  A.update(onWorkList,x,false);
               propagate(ys, WL)
           end

       fun id(NODE{id, ...}) = id
       fun add(x,y,WL) = 
       let val x = id x
           val y = id y
       in  addEdge(x,y);
           insertWL(y,WL)
       end
                        
       fun collect(NODE{kind=ref Undef, ...},_) = raise FSA
         | collect(x as NODE{kind=ref(Epsilon y), ...},WL) = add(x,y,WL) 
         | collect(x as NODE{kind=ref(BeginGroup(_,y)), id, ...},WL) = 
              singleton(id, add(x,y,WL))
         | collect(x as NODE{kind=ref(EndGroup(_,y)), id, ...},WL) = 
              singleton(id, add(x,y,WL))
         | collect(x as NODE{kind=ref(Or ys), ...},WL) = 
             foldr (fn (y, WL) => add(x,y,WL)) WL ys
         | collect(NODE{kind=ref(Accept _), id, ...},WL) = singleton(id,WL)
         | collect(NODE{kind=ref(Delta _), id, ...},WL) = singleton(id,WL)
         | collect(NODE{kind=ref(Error), id, ...},WL) = empty(id,WL)
       and singleton(id,WL)  = (insert(id, S.add(S.empty, id)); WL)
       and empty(id, WL) = (insert(id, S.empty); WL)

       val WL = H.fold collect [] nodes
   in  if DEBUG then print("Closure begin\n") else (); 
       if DEBUG then app print (map n2s (H.listItems nodes)) else ();
       if DEBUG then print ("|worklist|="^i2s(length WL)^"\n") else ();
       iter WL; 
       if DEBUG then print("Closure end\n") else ();
       ()
   end

   fun closure(NFA{closure, ...}) =
   let val lookup = H.lookup closure
   in  fn NODE{id, ...} => lookup id 
   end

   fun id(NODE{id, ...}) = id

   fun move (NFA{nodes, closure, ...}) =
   let val closure = H.lookup closure
       val lookup  = H.lookup nodes
       fun add(n, delta) = add'(lookup n, delta)
       and add'(NODE{kind=ref(Delta(c,s)), ...}, delta) =
           let val S'  = closure(id s)
               val S'' = getOpt(M.find(delta,c), S.empty)
           in  M.insert(delta, c, S.union(S',S''))
           end
         | add'(_, delta) = delta
   in  fn S => M.listItemsi(
                 S.foldr (fn (n,m) => S.foldr add m (closure n)) M.empty S)
   end

   fun acceptingAll (NFA{nodes, closure, ...}) =
   let val closure = H.lookup closure
       val lookup  = H.lookup nodes
       fun add(n, rules) = add'(lookup n, rules)
       and add'(NODE{kind=ref(Accept r), ...}, rules) = S.add(rules, r)
         | add'(_, delta) = delta
   in  S.foldr(fn (n,m) => S.foldr add m (closure n)) S.empty
   end

   fun accepting (NFA{nodes, closure, ...}) =
   let val closure = H.lookup closure
       val lookup  = H.lookup nodes
       fun add(n, r) = add'(lookup n, r)
       and add'(NODE{kind=ref(Accept r), ...}, NONE) = SOME r
         | add'(NODE{kind=ref(Accept r), ...}, SOME r') = SOME(Int.min(r,r'))
         | add'(_, r) = r
   in  S.foldr(fn (n,m) => S.foldr add m (closure n)) NONE 
   end

   fun groupings (NFA{nodes, closure, ...}) =
   let val closure = H.lookup closure
       val lookup  = H.lookup nodes
       fun add(n, g) = add'(lookup n, g)
       and add'(NODE{kind=ref(BeginGroup(g,_)), ...}, (x,y)) = (S.add(x,g),y)
         | add'(NODE{kind=ref(EndGroup(g,_)), ...}, (x,y)) = (x,S.add(y,g))
         | add'(_, g) = g
       fun get S = 
       let val (start, stop) = 
           S.foldr(fn (n,m) => S.foldr add m (closure n)) (S.empty,S.empty) S
       in  (S.listItems start, S.listItems stop) 
       end
   in  get
   end

end

end
