(*
 * Yet another dfa based pattern matching engine.
 * 
 * Features: 
 *  1. LexerEngine is a bit faster than DfaEngine in dfa construction.
 *  2. The full match tree is returned upon a successful match.
 *  3. The metacharacter ^ and $ are supported.
 *     However, backreferences and other zero-length assertions like \b 
 *     are unsupported.
 *
 * This version is specialized to the usual 8-bit characters.
 * Note: if you don't need the full match tree, use FastLexerEngine.
 *
 * IMPORTANT NOTE:
 * 
 *  The semantics of matchtrees is different than the other backtrack
 *  based engines.  In particular, if multiple alteratives match then *all*
 *  results are returned.  For example, if the pattern
 *
 *    /a(.*(cd)|(.+)d)e/
 *
 *  is applied on the text abcde,
 *  then the resulting match tree will look like this:
 *
 *  Match(SOME{pos=1,len=3},           -- the string bcd 
 *     [Match(SOME{pos=2,len=2},[]),   -- the string cd 
 *      Match(SOME{pos=1,len=2},[])])  -- the string bc 
 *
 *  Note the bindings are produced for both alternatives .*(cd) and (.+)d, 
 *  because the substring bcd matches both.  
 *
 *  -- Allen
 *)

local

   val DEBUG = false

in

structure LexerEngine : REGEXP_ENGINE =
struct

   structure R     = RegExpSyntax 
   structure Char  = R.Char
   structure Gen   = LexerGenerator(R)
   structure Lexer = LexerTables
   structure M     = MatchTree
   structure A     = Array
   structure IM    = IntRedBlackMap

   type regexp = Gen.Lexer.lexer

   fun compile re = Gen.compile [re]

   val error = 0 (* the error state is always 0 *)

   val i2s = Int.toString

   (* Create a new scanner *) 
   fun scanner(lexer as 
                  Lexer.LEXER{equiv, base, check, next, skels, numGroups, 
                              default, accept, groups, eof, 
                              start, start', ...},
               prefixOnly
              ) =
   let (* Mark the beginning/end of groups for a state *)
       fun mark(state,pos,s,starts,stops) =
       let fun openGroup(g,starts,stops) = 
               (if DEBUG then 
                    print("OPEN "^i2s g^" state "^i2s state^
                          " pos="^i2s pos^"\n")
                else ();
                (IM.insert(starts,g,(pos,s)), stops)
               )
           fun closeGroup(g,starts,stops) =
           let val SOME (pos',s) = IM.find(starts,g)
           in  if DEBUG then 
                  print("CLOSE "^i2s g^" state "^i2s state^
                        " pos="^i2s pos'^"-"^ i2s pos^"\n") else ();
               (starts, IM.insert(stops,g,(pos',pos,s)))
           end
           fun interp(actions, starts, stops) =
               case actions of
                 [] => (starts, stops)
               | Lexer.OPEN g::rest => 
                 let val (starts, stops) = openGroup(g, starts, stops)
                 in  interp(rest, starts, stops)
                 end
               | Lexer.CLOSE g::rest =>
                 let val (starts, stops) = closeGroup(g, starts, stops)
                 in  interp(rest, starts, stops)
                 end
               | Lexer.OPENCLOSE g::rest =>
                 case (IM.find(starts,g), IM.find(stops,g)) of
                   (NONE, NONE) => openCloseGroup(g, rest, starts, stops)
                 | (SOME _,NONE) => closeOpenGroup(g, rest, starts, stops)
                 | (NONE, SOME _) => openCloseGroup(g, rest, starts, stops)
                 | (SOME(p1, _), SOME(p2, _, _)) =>
                   if p1 = p2 then openCloseGroup(g, rest, starts, stops)
                   else closeOpenGroup(g, rest, starts, stops)
           and openCloseGroup(g, rest, starts, stops) =
               let val (starts, stops) = openGroup(g, starts, stops)
                   val (starts, stops) = closeGroup(g, starts, stops)
               in  interp(rest, starts, stops)
               end
           and closeOpenGroup(g, rest, starts, stops) = 
               let val (starts, stops) = closeGroup(g, starts, stops)
                   val (starts, stops) = openGroup(g, starts, stops)
               in  interp(rest, starts, stops)
               end
       in  interp(A.sub(groups, state), starts, stops)
       end
     

       (* Generate the match tree from a successful match *) 
       fun results(endGroups, rule, stream, endPos) = 
       let fun empty (Lexer.SKEL(g,sks)) = M.Match(NONE, map empty sks)
           fun walk (Lexer.SKEL(g,sks)) = 
               case IM.find(endGroups,g) of 
                 NONE => M.Match(NONE,map empty sks)
               | SOME(start,stop,s) => 
                  M.Match(SOME{pos=s,len=stop-start}, map walk sks)
       in  M.Match(SOME{pos=stream, len=endPos}, map walk(A.sub(skels,rule)))
       end

       (* This function decodes the transition tables *)
       fun go(s,c) =
       let val offset = A.sub(base,s) + c
       in  if A.sub(check,offset) = s 
           then A.sub(next,offset)
           else let val s' = A.sub(default,s) 
                in  if s' = error then error else go(s',c)
                end
       end

       val state0 = A.sub(start, 0)
       val state0' = A.sub(start', 0)
       val rule0 = A.sub(accept, state0)
       val rule0' = A.sub(accept, state0')

       fun doScan getc stream = 
       let fun scan(s0, r0, pos, stream) = 
           let 
               fun goto(state, pos, s, starts, stops,
                        rule_ok, pos_ok, s_ok, starts_ok, stops_ok) =
               let val (starts,stops) = mark(state, pos, s, starts, stops)
                   val rule = A.sub(accept, state)
                   val (rule_ok, pos_ok, s_ok, starts_ok, stops_ok) =
                       if rule >= 0 then
                          (if DEBUG then print("ACCEPT state "^i2s state^"\n") 
                           else ();
                           (rule, pos, s, starts, stops)
                          )
                       else
                          (rule_ok, pos_ok, s_ok, starts_ok, stops_ok) 
               in  case getc s of
                     NONE => 
                        endOfStream(state, pos, s, starts, stops,
                                    rule_ok, pos_ok, s_ok, stops_ok)
                   | SOME(c,s) =>
                     let val state = go(state, A.sub(equiv,Char.ord c))
                     in  if state = error 
                         then finish(rule_ok, pos_ok, s_ok, stops_ok) 
                         else  goto(state, pos+1, s, starts, stops,
                                    rule_ok, pos_ok, s_ok, starts_ok, stops_ok)
                         (*
                         let val rule = A.sub(accept,state)
                             val pos = pos+1
                         in  if rule >= 0 then
                               (if DEBUG then print "ACCEPT\n" else ();
                                goto(state, pos, s, starts, stops, 
                                     rule, pos, s, starts, stops)
                               )
                             else
                               goto(state, pos, s, starts, stops, 
                                    rule_ok, pos_ok, s_ok, starts_ok, stops_ok)
                         end
                         *)
                     end
               end
               and finish(~1, pos, s, endGroups) = 
                    if prefixOnly then NONE 
                    else if state0 = error then fastcontinue(pos, stream)
                         else continue(pos, stream)
                 | finish(rule, pos, s, endGroups) = 
                    SOME((results(endGroups, rule,stream,pos), rule), s)
    
               and endOfStream(state, pos, s, starts, stops, 
                               rule_ok, pos_ok, s_ok, stops_ok) =
                   let (* val (starts,stops) = mark(state, pos, s, starts, stops)
                       val rule = A.sub(accept, state)
                       val (rule_ok, pos_ok, s_ok, stops_ok) =
                           if rule >= 0 then
                              (rule, pos, s, stops)
                           else
                              (rule_ok, pos_ok, s_ok, stops_ok)  *)
                       val state = go(state,eof)
                   in  if state = error 
                       then finish(rule_ok, pos_ok, s_ok, stops_ok)
                       else 
                       let val rule = A.sub(accept,state)
                       in  if rule >= 0 then (* An accept state *)
                           let val (_,stops) = 
                               mark(state,pos,s,starts,stops)
                           in  if DEBUG then print "EOF ACCEPT\n" else ();
                               finish(rule, pos, s, stops)
                           end
                           else finish(rule_ok, pos_ok, s_ok, stops_ok)
                       end
                  end
    
               and continue(pos, s) = 
                   case getc s of 
                     NONE => NONE
                   | SOME(#"\n",s) => scan(state0', rule0', pos+1, s)
                   | SOME(_,s) => scan(state0, rule0, pos+1, s) 

               and fastcontinue(pos, s) = 
                   case getc s of 
                     NONE => NONE
                   | SOME(#"\n",s) => scan(state0',rule0', pos+1, s)
                   | SOME(_,s) => fastcontinue(pos+1, s) 
         
               (* 
               val (starts, stops) = mark(s0, pos, stream, IM.empty, IM.empty)
               val (starts_ok, stops_ok, pos_ok) =  
                    if r0 >= 0 then (starts, stops, pos)
                    else (IM.empty, IM.empty, ~1)
               *)
               val (starts, stops) = (IM.empty, IM.empty)
               val (starts_ok, stops_ok, pos_ok) = (IM.empty, IM.empty, ~1)
           in  goto(s0, pos, stream, starts, stops, 
                    r0, pos_ok, stream, starts_ok, stops_ok)
           end
       in  scan(state0', rule0', 0, stream)
       end
    
   in  doScan 
   end

   fun prefixOrFind(lexer,prefixOnly) = 
   let val scanner = scanner(lexer, prefixOnly)
       fun scan getc stream =
           case scanner getc stream of
             SOME((m, _), s) => SOME(m, s)
           | NONE => NONE
   in  scan
   end

   fun prefix lexer = prefixOrFind(lexer,true)
   fun find   lexer = prefixOrFind(lexer,false)

   fun match rules =
   let val lexer = Gen.compile (map (fn (re,_) => re) rules)
       val actions = Vector.fromList (map #2 rules)
       val scanner = scanner(lexer, true)
       fun matcher getc stream =
           case scanner getc stream of
              SOME((m, ruleno), s) => SOME((Vector.sub(actions,ruleno)) m,s)
           |  NONE                 => NONE
   in  matcher
   end

end (* LexerEngine *)
end (* local *)
