(*
 * Yet another lexer generator.  It supports the following features:
 *
 * 1. The meta characters ^ and $
 * 2. contexts
 * 3. wider characters than 8 bit (for future unicode support)
 * 4. perl style grouping constructs 
 *
 * IMPLEMENTATION NOTE: 
 *  1. Contexts and ^ patterns are handled by creating multiple start states.
 *  2. $ is handled by explicitly matching against an eof character in the
 *     lexer driver.
 *  3. Groupings are handled by creating begin group and end group states
 *     in the nfa.  These are compiled into a grouping table, which is
 *     a mapping from a dfa state to its sets of begin group and end group 
 *     actions.  The driver then takes care of keeping track of the groups.
 *
 * -- Allen Leung
 *)

local

   val DEBUG = false         (* for debugging messages only *)
   val VERBOSE = false       (* verbose debugging messages *)

   val MATCHTREES = true     (* should we enable the matchtrees recording
                              * feature in the generated lexer
                              *)

in

functor LexerGenerator(R : GENERALIZED_REGEXP_SYNTAX) : LEXER_GENERATOR =
struct

   structure R        = R                     (* The regexp syntax *)
   structure Char     = R.Char                (* Character representation *)
   structure Eqv      = EquivCharClassFn(R)   (* Equivalence class of chars *)
   structure Automata = DenseAutomataTable    (* How to encode automata *)
   structure Lexer    = LexerTables
   structure DA       = Automata.DynArray
   structure A        = Array               
   structure IS       = FSA.IntSet
   structure M        = RedBlackMapFn(type ord_key = FSA.IntSet.set
                                      val compare = FSA.IntSet.compare)

   type context = int
   datatype rule = 
      RULE of
        { contexts  : context list,
          regexp    : R.syntax,
          lookahead : R.syntax option
        }

   type state = int
   type symbol = int

   val i2s = Int.toString
   fun s2s s = "{"^IS.foldr(fn (i,"") => i2s i | (i,s) => i2s i^","^s) "" s^"}"

   val new = FSA.newNode 

   (* Strip away the groups *)
   fun strip(R.Group e) = strip e
     | strip(e as R.Char _) = e
     | strip(e as R.MatchSet _) = e
     | strip(e as R.NonmatchSet _) = e
     | strip(e as R.Begin) = e
     | strip(e as R.End) = e
     | strip(R.Interval(e,min,max)) = R.Interval(strip e,min,max)
     | strip(R.Star e) = R.Star(strip e)
     | strip(R.Plus e) = R.Plus(strip e)
     | strip(R.Option e) = R.Option(strip e)
     | strip(R.Alt es) = R.Alt(map strip es)
     | strip(R.Concat es) = R.Concat(map strip es)
     | strip(R.Guard(g,e)) = R.Guard(g,strip e)
     | strip(e as R.Ref _) = e
     | strip(R.Assign(x,f,e)) = R.Assign(x,f,strip e)
     | strip(e as R.Boundary _) = e

   (* Translate regexp into nfa *)
   fun buildRule {nfa, equiv, allClasses, matchtrees,
                  ruleno, rule=RULE{regexp, lookahead, contexts, ...},
                  startStates, startStates', groupIndex, eofOrd
                 } =
   let 
       (* The equivalence class of a character *)
       fun classOf c = A.sub(equiv,Char.ord c)

       (* The equivalence classes of a matchset *)
       val classOfSet = 
           R.CharSet.foldr (fn (c,S) => IS.add(S, classOf c)) IS.empty

       (* The equivalence classes of a non matchset *)
       fun classOfNonSet S = IS.difference(allClasses, classOfSet S)

       (* Build a nfa given 
        *    1. the preorder group index 
        *    2. the regexp
        *    3. the nfa node that terminates the current nfa
        * Returns:
        *    1. the final preorder group index
        *    2. the start node of the nfa
        *    3. a match tree skeleton
        *
        * Note that the preorder group index is ordered from right to left.
        * However, since it is only for internal use it is unimportant.
        * The only important thing is that the generated skeleton matches  
        * the structure of the structure properly and refers to the appropriate
        * group index.
        * 
        *)
       fun build(i, R.Group e, n) = 
           if MATCHTREES andalso matchtrees then
              let val (j, n, sk) = build(i+1, e, FSA.setEndGroup(new nfa, i, n))
              in  (j, FSA.setBeginGroup(new nfa, i, n), [Lexer.SKEL(i, sk)])
              end
           else build(i, e, n)
         | build(i, R.Char c, n) = (i, FSA.setDelta(new nfa, classOf c, n), [])
         | build(i, R.MatchSet s, n) = buildSet(i, classOfSet s, n)
         | build(i, R.NonmatchSet s, n) = buildSet(i, classOfNonSet s, n)
         | build(i, R.Concat es, n) = buildList(i, es, n)
         | build(i, R.Plus e, n) = 
           let val n1 = new nfa
               val (i, n2, sk) = build(i, e, n1)
           in  FSA.setOr(n1, [n, n2]);
               (i, n2, sk)
           end
         | build(i, R.Star e, n) = 
           let val n1 = new nfa
               val (i, n2, sk) = build(i, e, n1)
           in  (i, FSA.setOr(n1, [n, n2]), sk)
           end
         | build(i, R.Option e, n) = 
           let val (i, n', sk) = build(i, e, n)
           in  (i, FSA.setOr(new nfa,[n, n']), sk)
           end
         | build(i, R.Alt [e], n) = build(i, e, n)
         | build(i, R.Alt es, n) = 
           let fun iter(i, [], ns, sks) = (i, ns, List.concat(rev sks))
                 | iter(i, e::es, ns, sks) = 
                   let val (i, n', sk) = build(i, e, n)
                   in  iter(i, es, n'::ns, sk::sks)
                   end
               val (i, ns, sk) = iter(i, es, [], [])
           in  (i, FSA.setOr(new nfa, ns), sk)
           end

           (* Translate intervals to the appropriate things *)
         | build(i, R.Interval(e,0,NONE), n) = build(i, R.Star e, n)
         | build(i, R.Interval(e,1,NONE), n) = build(i, R.Plus e, n)
         | build(i, R.Interval(e,0,SOME 1), n) = build(i, R.Option e, n)
         | build(i, R.Interval(e,1,SOME 1), n) = build(i, e, n)
         | build(i, R.Interval(e,min,NONE), n) = 
              build(i, R.Concat[strip e, R.Interval(e, min-1, NONE)], n)
         | build(i, R.Interval(e,0,SOME max), n) = (* XXX *)
              build(i, R.Option(R.Concat[strip e, 
                                         R.Interval(e, 0, SOME(max-1))]), n)
         | build(i, R.Interval(e,min,SOME max), n) = 
              if min > max then 
                  (* can never match; but we'll still have to generate
                   * the match skeletons
                   *)
                 let val (i, _, sk) = build(i, e, n)
                 in  (i, n, sk)
                 end
              else build(i, R.Concat[strip e, 
                                     R.Interval(e, min-1, SOME(max-1))], n)

         | build(i, R.End, n) = 
              if false then (i, n, [])
              else (i, FSA.setDelta(new nfa, eofOrd, n), []) 
         | build _ = raise R.CannotCompile

           (*
            * Concatenate the nfas
            *)
       and buildList(i, [], n) = (i, n, [])
         | buildList(i, [e], n) = build(i, e, n)
         | buildList(i, es, n) = 
           let fun iter(i, [], n) = (i, n, [])
                 | iter(i, e::es, n) = 
                   let val (i, n, sks) = iter(i, es, n)
                       val (i, n, sk)  = build(i, e, n)
                   in  (i, n, sk@sks)
                   end
           in  iter(i, es, n)
           end

       and buildSet(i, set, n) =
           let val ns = IS.foldr (fn (c,ns) =>
                           FSA.setDelta(new nfa,c,n)::ns) [] set 
           in  (i, FSA.setOr(new nfa, ns), [])
           end


       fun preprocess (R.Concat[x]) = preprocess x
         | preprocess (R.Alt[x]) = preprocess x
         | preprocess (R.Begin) = (true, R.Concat[])
         | preprocess (R.Concat(R.Begin::es)) = (true, R.Concat es)
         | preprocess e = (false, e)
 
       val (isBeginLine,regexp) = preprocess regexp
       val final = FSA.setAccept(new nfa, ruleno)
       val (i, node, skel) = build(groupIndex, regexp, final)

       (* Determine which start states can branch to this nfa *)
       fun addToContexts tbl c = 
           A.update(tbl,c, FSA.setOr(new nfa,[node, A.sub(tbl, c)]))
       val contexts = case contexts of [] => [0] | _ => contexts
   in  if isBeginLine then () else app (addToContexts startStates) contexts;
       app (addToContexts startStates') contexts;
       (i, skel)
   end

   (*
    * Main entry point of the lexer generator
    *)
   fun compile' {matchtrees, numContexts} rules = 
   let 

       (* Compute the equivalence classes from all the regexps
        * This is NOT the most general partitioning, but rather
        * a quick and easy approximation.  For most purposes this
        * works very well.
        *) 
       val eofChar    = Char.maxOrd + 1 
       val equivclass = Eqv.Eqv.new{min=0, max=eofChar}
       val () = app (fn RULE{regexp, lookahead, ...} =>
                          (Eqv.partition equivclass regexp;
                           case lookahead of
                             SOME re => Eqv.partition equivclass re
                           | NONE => ())) rules
       val ()                    = Eqv.Eqv.single equivclass eofChar
       val {mapping, numclasses} = Eqv.Eqv.mapping equivclass
       val eofOrd                = A.sub(mapping,eofChar)
       val allClasses = 
           let fun iter(~1,S) = S
                 | iter(i, S) = iter(i-1,if i = eofOrd then S else IS.add(S,i))
           in  iter(numclasses, IS.empty) end

       (* Allocate the nfa *)
       val nfa = FSA.nfa{min=0, max=eofChar}

       (* Allocate the start state tables *)
       val numStartStates = numContexts +1
       val start        = A.array(numStartStates, Automata.error)
       val start'       = A.array(numStartStates, Automata.error)
       val fail         = FSA.setError(new nfa)
       val startStates  = A.array(numStartStates, fail)
       val startStates' = A.array(numStartStates, fail)

       (* Translate the rules into an nfa *)
       fun buildRules(i, ruleno, [], skels) = (i, rev skels)
         | buildRules(i, ruleno, rule::rules, skels) = 
           let val (i,skel) = 
                   buildRule{nfa=nfa, equiv=mapping, 
                             allClasses=allClasses, ruleno=ruleno,
                             rule=rule, matchtrees=matchtrees,
                             startStates=startStates, 
                             startStates'=startStates', groupIndex=i,
                             eofOrd=eofOrd
                            }
           in  buildRules(i, ruleno+1, rules, skel::skels)
           end

       val (numberOfGroups, skels) = buildRules(0, 0, rules, [])

       (* Compute the closure *)
       val ()   = FSA.computeClosure nfa
       
       (* Create the dfa *)
       val automata      = Automata.new{min=0, max=numclasses-1}
       val addState      = Automata.addState automata
       val move          = FSA.move nfa
       val accepting     = FSA.accepting nfa 
       val groupings     = FSA.groupings nfa 
       val groups        = DA.array(0,[])
       val accept        = DA.array(10,~1)
       val numberOfRules = length rules
       val skels = if MATCHTREES andalso matchtrees then A.fromList skels 
                   else A.array(0,[])

       fun compileGroups(openings, closings) =
       let fun rearrange([], []) = []
             | rearrange(a, []) = map Lexer.OPEN a
             | rearrange([], b) = map Lexer.CLOSE b
             | rearrange(a as x::xs, b as y::ys) = 
               if x = y then Lexer.OPENCLOSE x::rearrange(xs, ys)
               else if x < y then Lexer.OPEN x::rearrange(xs, b)
               else Lexer.CLOSE y::rearrange(a, ys)
       in  rearrange(openings,closings)
       end 

       (* Create a new state numbered n *)
       fun newState(n, S, stateMap) = 
       let val stateMap = M.insert(stateMap, S, n)
       in  case accepting S of
             NONE => ()
           | SOME r => DA.update(accept, n, r);
           if MATCHTREES andalso matchtrees 
           then DA.update(groups, n, compileGroups(groupings S)) else ();
           if DEBUG andalso VERBOSE then
              print("New dfa state "^i2s n^"="^s2s S^"\n")
           else ();
           stateMap
       end

       (* nfa -> dfa conversion *)
       fun nfa2dfa(n, [], stateMap) = ()
         | nfa2dfa(n, (s,S)::pending, stateMap) = 
           let fun trans([], n, stateMap, pending, tr) =
                     (n, stateMap, pending, tr)
                 | trans((c,S')::deltas, n, stateMap, pending, tr) = 
                   let val (n,s',stateMap,pending) = 
                           case M.find(stateMap,S') of
                             SOME s' => (n,s',stateMap,pending)
                           | NONE => (n+1,n,newState(n,S',stateMap),
                                      (n,S')::pending)
                   in  trans(deltas, n, stateMap, pending, (c,s')::tr)
                   end
               val (n, stateMap, pending, transitions) = 
                    trans(move S, n, stateMap, pending, [])
           in  addState (s, transitions);
               nfa2dfa(n, pending, stateMap)
           end
  
        val () = if DEBUG andalso VERBOSE then 
                    print "Creating start states\n" else ()

        fun makeStartStates(tbl,tbl',n, ~1,stateMap, WL) = (n, stateMap, WL)
          | makeStartStates(tbl,tbl',n, c, stateMap, WL) = 
            let val startNode = A.sub(tbl,c)
                val S         = FSA.closure nfa startNode
                val (s, n, stateMap, WL) = 
                       case M.find(stateMap,S) of
                         SOME s => (s, n, stateMap, WL)
                       | NONE   => (n, n+1, newState(n, S, stateMap), (n,S)::WL)
            in  A.update(tbl', c, s);
                makeStartStates(tbl,tbl',n,c-1,stateMap, WL)
            end
         
        (* Create error state *) 
        val stateMap = newState(0,IS.empty,M.empty)
        val (n,stateMap,WL) =
               makeStartStates(startStates,start,1,numContexts,stateMap,[])
        val (n,stateMap,WL) =
               makeStartStates(startStates',start',n,numContexts,stateMap,WL)
        val () = if DEBUG andalso VERBOSE then print "NFA->DFA start\n" else ()
        val () = nfa2dfa(n, WL, stateMap)
        val () = if DEBUG andalso VERBOSE then print "NFA->DFA end\n" else ()
       
       (* Finish compression *)    
       val () = Automata.compress automata

       (* The tables *)    
       val base    = Automata.getarr automata "base"
       val check   = Automata.getarr automata "check"
       val next    = Automata.getarr automata "next"
       val default = Automata.getarr automata "default"
       val numStates = Automata.stateCount automata
       val ()      = DA.setBound(accept, numStates)
       val ()      = DA.setBound(groups, numStates)
       val accept  = DA.toArray accept
       val groups  = DA.toArray groups

       val () =
           if DEBUG then
              (
               print ("equivclasses = "^i2s numclasses^"\n");
               print ("states = "^i2s numStates^"\n");
               print ("numberOfGroups = "^i2s numberOfGroups^"\n");
               print ("|base|    = "^i2s(A.length base)^"\n");
               print ("|check|   = "^i2s(A.length check)^"\n");
               print ("|next|    = "^i2s(A.length next)^"\n");
               print ("|default| = "^i2s(A.length default)^"\n");
               print ("|equiv|   = "^i2s(A.length mapping)^"\n");
               print ("|skels|   = "^i2s(A.length skels)^"\n");
               print ("|accept|  = "^i2s(A.length accept)^"\n");
               print ("|groups|  = "^i2s(A.length groups)^"\n");
               print ("|start|   = "^i2s(A.length start)^"\n");
               print ("|start'|  = "^i2s(A.length start')^"\n");
               ()
              )
           else ()

   in  Lexer.LEXER
       { base        = base,
         check       = check,
         next        = next,
         default     = default,
         equiv       = mapping,
         accept      = accept,
         groups      = groups, 
         skels       = skels,
         start       = start,
         start'      = start',
         numGroups   = numberOfGroups,
         numStates   = numStates,
         eof         = eofOrd
       }
   end

   fun compile'' matchtrees re = 
       compile'{matchtrees=matchtrees, numContexts=0} 
               (map (fn re => RULE{regexp=re,lookahead=NONE,contexts=[]}) re) 

   val compile  = compile'' true
   val compileWithoutMatchtrees = compile'' false
end

end (* local *)
